from collections import namedtuple
import random as rd
import math

Color = namedtuple("Color", ["r", "g", "b"])

BLUE = Color(0, 0, 255)
RED = Color(255, 0, 0)
GREEN = Color(0, 255, 0)
YELLOW = Color(255, 255, 0)
BLACK = Color(0, 0, 0)
WHITE = Color(255, 255, 255)
GREY = Color(100, 100, 100)
PURPLE = Color(100, 0, 100)
ORANGE = Color(255, 165, 0)
BROWN = Color(160, 82, 45)
HALFGREY = Color(50, 50, 50)
DARKGREY = Color(20, 20, 20)
DARKRED = Color(10, 10, 10)
DARKGREEN = Color(0, 100, 0)
DARKBLUE = Color(0, 0, 100)
LIGHTRED = Color(255, 200, 200)
LIGHTGREEN = Color(200, 255, 200)
LIGHTBLUE = Color(200, 200, 255)
LIGHTBROWN = Color(229, 219, 222)
LIGHTGREY = Color(200, 200, 200)
BEIGE = Color(199, 175, 138)

sigmoid = s = lambda x: 1 / (1 + math.exp(-x))
reverse_sigmoid = lambda x: math.log(x / (1 - x))
# r=reverse_sigmoid()

bijection = lambda x, e, s: (x - e[0]) / (e[1] - e[0]) * (s[1] - s[0]) + s[0]

random = lambda: Color(*[rd.randint(0, 255) for i in range(3)])
reverse = lambda color: Color(*[255 - c for c in color])
darken = lambda color, n=0: Color(*[int(c * sigmoid(n / 10)) for c in color])
lighten = lambda color, n=0: Color(
    *[int(255 - (255 - c) * sigmoid(n / 10)) for c in color]
)
mix = lambda cl1, cl2: Color(*[(c1 + c2) // 2 for (c1, c2) in zip(cl1, cl2)])
substract = lambda cl1, cl2: Color(
    *[max(min(2 * c1 - c2, 255), 0) for (c1, c2) in zip(cl1, cl2)]
)
increase = lambda color, n=2: Color(
    *[int(255 * math.exp(n * math.log(c / 255))) for c in color]
)


def nuance(color1, color2, t, p=1 / 2):
    """Return a color between the two depending on the degree."""
    return Color(
        *[c1 * (1 - t ** p) + c2 * (t ** p) for (c1, c2) in zip(color1, color2)]
    )


def nuances(colors, t):
    """Does an interpolation between all the colors in order."""
    n = len(colors)
    int(t * n)


def ponderNuances(colors, t):
    """Does an interpolation between all the colors in order."""
    n = len(colors)
    int(t * n)


def getFromWavelength(wavelength):
    """Return a color using wavelength."""
    gamma, max_intensity = 0.80, 255

    def adjust(color, factor):
        if color == 0:
            return 0
        else:
            return round(max_intensity * pow(color * factor, gamma))

    if 380 <= wavelength <= 440:
        r, g, b = -(wavelength - 440) / (440 - 380), 0, 1
    elif 440 <= wavelength <= 490:
        r, g, b = 0, (wavelength - 440) / (490 - 440), 1
    elif 490 <= wavelength <= 510:
        r, g, b = 0, 1, -(wavelength - 510) / (510 - 490)
    elif 510 <= wavelength <= 580:
        r, g, b = (wavelength - 510) / (580 - 510), 1, 0
    elif 580 <= wavelength <= 645:
        r, g, b = 1, -(wavelength - 645) / (645 - 580), 0
    elif 645 <= wavelength <= 780:
        r, g, b = 1, 0, 0
    else:
        r, g, b = 0, 0, 0
    if 380 <= wavelength <= 420:
        factor = 0.3 + 0.7 * (wavelength - 380) / (420 - 380)
    elif 420 <= wavelength <= 701:
        factor = 1
    elif 701 <= wavelength <= 780:
        factor = 0.3 + 0.7 * (780 - wavelength) / (780 - 700)
    else:
        factor = 0
    r, g, b = adjust(r, factor), adjust(g, factor), adjust(b, factor)
    return Color(r, g, b)


def colorRange(t):
    bt = bijection(t, [0, 1], [380, 780])
    return getFromWavelength(bt)


if __name__ == "__main__":
    print(darken(RED, 10))
    print(mix(YELLOW, RED))
    print(reverse(LIGHTBROWN))
    print(substract(LIGHTBROWN, ORANGE))
    print(increase(LIGHTBROWN))
    print(nuance(YELLOW, RED, 10))
    print("colorRange(0.5):", colorRange(0.1))
    print(BLUE.r)
    print("mycolors imported")



"""
The goal of this program is to make an interpolation of the points of any
curve in order to make it a parametric function.


There are 2 main types of interpolations:
- Polynomial Interpolation:
- Bezier Interpolation:
"""

# Polynomial Interpolation
from .polynomial import Polynomial

# Pseudo Bezier Interpolation
from .curves import BezierCurve
from .abstract import Line, Point

# additional imports for showing the interpolations more easily
from .curves import Trajectory
from . import colors

# Deprectated function
def directInterpolation(points, t):
    x = [pt[0] for pt in points]
    y = [pt[1] for pt in points]
    plx = Polynomial.createFromInterpolation(range(len(x)), x)
    ply = Polynomial.createFromInterpolation(range(len(y)), y)
    l = len(pts)
    mst = 100

    # Creation of the points
    # pts=[(random.uniform(xmin,xmax),random.uniform(ymin,ymax)) for i in range(n)]

    # # """ Tests:
    # # Calculations for n points in dimension 2:
    #
    # x=[pt[0] for pt in pts]
    # y=[pt[1] for pt in pts]
    # plx=Polynomial.createFromInterpolation(range(len(x)),x)
    # ply=Polynomial.createFromInterpolation(range(len(y)),y)
    # print(plx)
    # print(ply)
    # l=len(pts)
    # mst=100
    # npts=[(plx(l*st/mst),ply(l*st/mst)) for st in range(mst)]
    # grapher=Grapher(context,functions=[plx,ply])
    # """


class Interpolation:
    """Base class of all interpolations."""

    def __init__(self, points, create=True):
        """Create the interpolation using a list of n-dimensional points."""
        self.points = points
        if create:
            self.create()

    def create(self):
        """Create the necessary attributes for the interpolation.
        This method is to be overloaded."""
        pass

    def sample(self, n):
        """Make a sample of the interpolation for n points."""
        m = n - 1
        return [self(t / m) for t in range(n)]

    def __str__(self):
        """Return the string representation of the interpolation."""
        return "Itp(" + ",".join(map(str, self.points)) + ")"


class ExtraInterpolation(Interpolation):
    """Extra features for interpolations."""

    def getSegments(self):
        """Return the segments formed by the points."""
        return [
            Segment(self.points[i], self.points[i + 1])
            for i in range(len(self.points) - 1)
        ]

    def getTrajectory(self):
        """Return the trajectory associated with the points of the interpolation."""
        return Trajectory.createFromTuples(self.points)

    def getPointsType(self):
        """Return the points under a point type instead of just a tuple type."""
        return [Point(*p) for p in self.points]

    def getPoint(self, t, **kwargs):
        """Return the point of position t in the parametric function."""
        return Point(*self(t), **kwargs)

    segments = property(getSegments)
    trajectory = property(getTrajectory)
    points_type = property(getPointsType)


class VisualInterpolation(Interpolation):
    """Visual features for interpolation, this class allows to show an
    interpolation on a graphical context."""

    def __init__(
        self,
        points,
        color=colors.RED,
        point_color=colors.LIGHTRED,
        trajectory_color=colors.GREEN,
        **kwargs
    ):
        """Create a visual interpolation."""
        self.color = color
        self.point_color = point_color
        self.trajectory_color = trajectory_color
        Interpolation.__init__(self, points, **kwargs)

    def show(self, context, n, color=None):
        """Show the polynomial interpolation by sampling the function using:
        -n: the number of points
        -context:the context in which the interpolation is shown."""
        if not color:
            color = self.color
        points = self.sample(n)
        context.draw.lines(context.screen, color, points, width=1, connected=False)

    def showTrajectory(self, context, color=None):
        """Show the trajectory defined by the points of the interpolation."""
        if not color:
            color = self.trajectory_color
        Trajectory.createFromTuples(self.points).show(context, color=color)


class PolynomialInterpolation(VisualInterpolation, ExtraInterpolation):
    """Make an interpolation using the lagrangian polynomial interpolator
    over list of n-dimensional points."""

    def create(self):
        """Create the components useful for the interpolation which are the
        split components and the lagrangian polynomials."""
        self.createComponents()
        self.createPolynomials()

    def createComponents(self):
        """Split its components of each point in separate lists."""
        self.components = [
            [pt[i] for pt in self.points] for i in range(len(self.points[0]))
        ]

    def createPolynomials(self):
        """Create its polynomials using its components."""
        lpts = len(self.points)
        self.polynomials = [
            Polynomial.createFromInterpolation(range(lpts), c) for c in self.components
        ]

    def __call__(self, t):
        """Evaluate the parametric interpolation for the input 't' between 0 and 1."""
        lcps = len(self.components)
        lpts = len(self.points) - 1
        return tuple([self.polynomials[i](lpts * t) for i in range(lcps)])


class PseudoBezierInterpolation(VisualInterpolation):
    """Make an interpolation using the bezier interpolation over a list of
    n-dimensional points."""

    def create(self):
        """Create the lines, the intruding points and the bezier curves."""
        self.createLines()
        self.createIntrudingPoints()
        self.createBezierCurves()

    def createLines(self):
        """Create the lines of the duos of consecutive points."""
        l = len(self.points)
        lines = [
            Line.createFromTwoPoints(self.points[i], self.points[i + 1])
            for i in range(l - 1)
        ]

    def createIntrudingPoints(self):
        """Create the points of intersection formed by some of the lines."""
        l = len(lines)
        intruding_points = [lines[i].crossLine(lines[i + 1]) for i in range(l - 1)]

    def __call__(self, t):
        """Evaluate the parametric interpolation for the input 't' between 0 and 1."""
        l = len(lines)
        intruding_points = [lines[i].crossLine(lines[i + 1]) for i in range(l - 1)]
        bezier_curves = []
        for i in range(l):
            bezier_points = [
                self.points[i],
                self.intruding_points[i],
                self.points[i + 1],
            ]
            b = BezierCurve(bezier_points)
            bezier_curves.append(b)


if __name__ == "__main__":
    # To show the points
    from mycontext import Context
    import colors

    # To create the points
    import random

    context = Context(name="Interpolation Demonstration")

    # Parameters
    h = 1
    w = 1
    xmin, xmax, ymin, ymax = [-w, w, -h, h]  # corners
    l = 5  # number of points
    n = 0  # number of steps
    m = 100  # max number of steps

    # Final version
    # pts=[Point.random() for i in range(10)]
    pts = [Point(2 * x, random.randint(-5, 5)) for x in range(l)]
    interpolation = PolynomialInterpolation([p.components for p in pts])
    interpolation.color
    npts = interpolation.sample(200)

    # Setting the console
    context.console.append(
        [
            "interpolation: " + str(interpolation),
            "trajectory: " + str(interpolation.trajectory),
        ]
    )

    # Main loop of the context
    while context.open:
        # Update the context
        context.check()
        context.control()
        context.clear()
        context.show()

        # Update data components
        Point.turnPoints([1 / 100 / l for i in range(l)], pts)
        # trajectory=Trajectory(pts,colors.GREEN)
        interpolation = PolynomialInterpolation([p.components for p in pts])
        npts = interpolation.sample(200)  # Sample 200 points by interpolation

        # Additional features
        n = (n + 1) % (m + 1)
        c1 = interpolation(n / m)
        pt1 = Point(*c1, radius=0.1, color=colors.lighten(colors.RED, 2), fill=True)
        c2 = interpolation.trajectory(n / m)
        pt2 = Point(*c2, radius=0.1, color=colors.lighten(colors.GREEN, 2), fill=True)

        # Show visual components
        pt1.show(context)
        pt2.show(context)
        interpolation.show(context, 200)
        interpolation.showTrajectory(context)

        # Console
        context.showConsole()

        # Flip the context
        context.flip()



from .plane import Plane
from .window import Window
from math import sqrt
from . import colors

"""
Structure:

Plane:
- init(name,size,fullscreen,theme,view)
- show(view) ?
- getFromScreen(position)
- getToScreen(position)



Draw:
-draw.circle(color,position,radius,fill)
-draw.rect(color,position,size,fill)
-draw.line(color,start_position,end_position,width)
-draw.lines(color,connected,positions,width)


Client:
-map.draw.rect(color,position,fill)
-map.draw.line(color,start_position)
"""


class Draw:
    def __init__(self, plane=Plane(), window=None, **kwargs):
        if window:
            self.window = window
        else:
            window = Window(**kwargs)
        self.plane = plane
        self.window = window
        self.window.text_size = 30

    def rect(self, screen, color, rect, fill=False, conversion=True):
        if conversion:
            position = rect[:2]
            size = rect[2:]
            sx, sy = size
            position = self.plane.getToScreen(position, self.window)
            ux, uy = self.plane.units
            size = [sx * ux + 1, sy * uy + 1]
            rect = position + size
        self.window.draw.rect(screen, color, rect, not (fill))

    def ellipse(self, screen, color, rect, fill=False):
        coordonnates = Plane.getCoordonnatesFromRect(rect)
        x, y, sx, sy = coordonnates
        py, px = self.plane.getToScreen((x, y), self.window)
        psx, psy = self.plane.getToScreen((sx, sy), self.window)
        pmx, pmy, pMx, pMy = Plane.getRectFromCoordonnates([px, py, psx, psy])
        rect = [pmx, pmy, pMx, pMy]
        self.window.draw.ellipse(screen, color, rect, fill)

    def circle(self, screen, color, position, radius, fill=False, conversion=True):
        # self.ellipse(screen,color,position+[radius,radius],fill)

        # rx,ry=self.plane.getToScreen([radius,radius],self.window)
        # rect=self.plane.getRectFromScreen(position+[radius,radius])
        # self.window.draw.ellipse(screen,color,rect,width)
        # Need to implement ellipses
        # r=int(radius/self.plane.units[0])
        # r,ry=self.plane.getToScreen([radius,radius],self.window)
        position = self.plane.getToScreen(position, self.window)
        x, y = position
        if conversion:
            rx, ry = [radius * self.plane.units[i] for i in range(2)]
            radius = int((rx + ry) / 2)
        if radius < 1:
            radius = 1
        # r=0.1
        self.window.draw.circle(screen, color, position, radius, not (fill))

    def square(self, screen, color, position, side_size, fill=False):
        position = self.plane.getToScreen(position, self.window)
        size = self.plane.getToScreen([side_size, side_size], self.window)
        self.window.draw.rect(screen, color, position + size, not fill)

    def polygon(
        self, screen, color, positions, fill=False
    ):  # No clue of what i'm doing to do here.
        screen_positions = self.plane.getAllToScreen(positions, self.window)
        self.window.draw.polygon(screen, color, screen_positions, fill)

    def line(
        self, screen, color, start_position, end_position, width=1, conversion=True
    ):
        if conversion:
            start_position = self.plane.getToScreen(start_position, self.window)
            end_position = self.plane.getToScreen(end_position, self.window)
        self.window.draw.line(screen, color, start_position, end_position, width)

    def lines(self, screen, color, positions, connected=True, width=1, conversion=True):
        if conversion:
            positions = self.plane.getAllToScreen(positions, self.window)
        self.window.draw.lines(screen, color, connected, positions, width)

    def arc(self, screen, color, rect, start_angle, stop_angle, width=1):
        position = rect[:2]
        size = rect[2:]
        position = self.plane.getToScreen(position, self.window)
        size = self.plane.getToScreen(size, self.window)
        self.window.draw.arc(screen, color, rect, start_angle, stop_angle, width)

    def point(self, screen, color, position, radius=1, **kwargs):
        self.circle(screen, color, position, radius, **kwargs)

    def complex(self, screen, color, z, radius=1, **kwargs):
        self.point(screen, color, (z.real, z.imag), radius, **kwargs)

    def image(self, screen, image, position, smax=1, corners=None):
        w, h = image.get_size()
        ux, uy = self.plane.units
        if corners is not None:
            x, y, sx, sy = corners
            rx, ry = self.plane.getToScreen([x, y], self.window)
            rsx = int(sx * ux)
            rsy = int(sy * uy)
            image = self.window.scale(image, (rsx, rsy))
        else:
            x, y = position
            rx, ry = self.plane.getToScreen([x, y], self.window)
            m = max(w, h)
            if w > h:
                sx = smax
                sy = h / w * smax
            else:
                sy = smax
                sx = w / h * smax
            rsx = int(sx * ux)
            rsy = int(sy * uy)
            image = self.window.scale(image, (rsx, rsy))
        self.window.screen.blit(image, (rx, ry))

    def print(
        self,
        text,
        position=None,
        size=None,
        color=colors.WHITE,
        font=None,
        conversion=True,
    ):
        """Print a text the window's screen using text and position and optional
        color, pygame font and conversion."""
        if len(color) == 4:
            color = self.applyTransparency(color)
        if conversion:
            if not position:
                position = (0, 0)
            if not size:
                size = 1
            position = self.plane.getToScreen(position, self.window)
            ux, uy = self.plane.units
            size = int(size * ux / 50)
        else:
            if not position:
                position = (10, 10)
            if not size:
                size = 20
        self.window.print(text, position, size, color, font)

    def applyTransparency(self, color):
        """Change the color to apply some transparency by merging it to the color
        of the background. This technique is more of a rough bug fix for pygame
        than a necessary function."""
        a = color[3]
        c1 = color[:3]
        c2 = self.plane.theme["background"]
        return colors.nuance(c1, c2, a)

    def show(self):
        self.plane.showGrid(self.window)
        self.window.flip()

    def clear(self, **kwargs):
        self.plane.clear(self.window, **kwargs)

    def check(self):
        self.window.check()

    def control(self):
        self.plane.control(self.window)


if __name__ == "__main__":
    window = Window("mydraw")
    plane = Plane()
    draw = Draw(plane, window)

    while draw.window.open:
        draw.check()
        draw.control()
        draw.clear()
        draw.line(draw.window.screen, colors.WHITE, (3, -5), (-2, 6))
        draw.show()



from .abstract import Point, Segment, Form, Line

import random
from . import colors
import itertools


class Trajectory:
    @classmethod
    def random(cls, n=5, **kwargs):
        """Create a random trajectory of n points."""
        points = [Point.random() for i in range(n)]
        return cls(points, **kwargs)

    @classmethod
    def createFromTuples(cls, tuples, **kwargs):
        """Create a trajectory using tuples and optional arguments."""
        pts = [Point(*t) for t in tuples]
        return cls(pts, **kwargs)

    def __init__(self, points, segment_color=colors.WHITE, point_color=colors.WHITE):
        """Create a trajectory using the list of points."""
        self.points = points
        self.segment_color = segment_color
        self.point_color = point_color

    def __call__(self, t):
        """Evaluate the trajectory by t between 0 and 1."""
        if len(self.points) <= 1:
            raise Exception(
                "It is impossible to interpolate without at least 2 points."
            )
        length = self.length
        ns = len(self.segments)
        tsl = 0
        for i in range(ns):
            s = Segment(self.points[i], self.points[i + 1])
            if t * length - tsl <= s.length:
                st = (t * length - tsl) / s.length
                return tuple(s(st).components)
            tsl += s.length
        return tuple(s(1))  # Its a fix that isn't always true

    def __str__(self):
        """Return the string representation of a trajectory."""
        return "tj(" + ",".join(map(str, self.points)) + ")"

    def show(self, surface, **kwargs):
        """Show the trajectory on the surface."""
        self.showPoints(surface, **kwargs)
        self.showSegments(surface, **kwargs)

    def showPoints(self, surface, color=None):
        """Show the point on the surface."""
        if not color:
            color = self.point_color
        for point in self.points:
            point.show(surface, color=color)

    def showSegments(self, surface, color=None):
        """Show the segments on the surface."""
        if not color:
            color = self.segment_color
        for segment in self.segments:
            segment.show(surface, color=color)

    def getSegments(self):
        """Return the segments of the trajectory of the curve."""
        return [
            Segment(self.points[i], self.points[i + 1])
            for i in range(len(self.points) - 1)
        ]

    def getLength(self):
        """Return the length of the trajectory."""
        return sum([segment.length for segment in self.segments])

    def sample(self, n, include=True):
        """Sample n points of the trajectory at equal distance.
        It is also possible to include the last one if wanted."""
        return [self(i / n) for i in range(n + int(include))]

    def fastSample(self, n, include=True):
        """Sample n points of the trajectory at equal distance but fast.
        It is also possible to include the last one if wanted."""
        length = self.length
        ns = len(self.segments)
        k = 0
        l = []
        tsl = 0
        for i in range(ns):
            s = Segment(self.points[i], self.points[i + 1])
            if t * length - tsl <= s.length:
                st = (t * length - tsl) / s.length
                l.append(tuple(s(st).components))
                k += 1
                t = k / n
            tsl += s.length
        if include:
            l.append(self.points[-1])
        return l

    def sampleSegments(self, n, include=True):
        """Sample n points for each segment of the trajectory."""
        points = []
        segments = self.segments
        for i in range(len(segments)):
            points += segments[i].sample(n, include=False)
        if include and len(self.points) > 1:
            points.append(self.points[-1])
        return points

    def getCenter(self):
        """Return the center (or middle) of the trajectory."""
        return self(1 / 2)

    def setCenter(self):
        """Set the center (or middle) of the trajectory."""
        center = self.center
        for n in range(len(self.points)):
            self.points[n] -= center

    segments = property(getSegments)
    length = property(getLength)
    center = property(getCenter)


class BezierCurve:
    def __init__(self, points, point_color=colors.WHITE, segment_color=colors.WHITE):
        """Create a bezier curve object using 3 points."""
        self.points = points
        self.segment_color = segment_color
        self.point_color = point_color

    def __call__(self, t):
        """Return the point."""
        points = self.points
        if len(points) == 0:
            return None
        elif len(points) == 1:
            return points[0]
        elif len(points) == 2:
            segment = Segment(points[0], points[1])
            point = segment.center()
            return point
        elif len(points) == 3:
            p1, p2, p3 = points
            s1 = Segment(p1, p2)
            s2 = Segment(p2, p3)
            ps1 = s1(t)
            ps2 = s2(t)
            s = Segment(ps1, ps2)
            return s(t)
        else:
            while len(points) > 3:
                segments = [
                    Segment(points[i], points[i + 1]) for i in range(len(points) - 1)
                ]
                points = [segment(t) for segment in segments]
            b = BezierCurve(points)
            return b(t)

    def show(self, surface, p=50):
        """Show the bezier curve on the surface."""
        points = [self(i / p) for i in range(p + 1)]
        segments = [Segment(points[i], points[i + 1]) for i in range(p)]
        self.showPoints(surface, points)
        self.showSegments(surface, segments)

    def getConstruction(self, t):
        """Return the construction segments of the form."""
        points = self.points
        construction = []
        while len(points) >= 2:
            segments = [
                Segment(points[i], points[i + 1]) for i in range(len(points) - 1)
            ]
            points = [segment(t) for segment in segments]
            construction.append(segments)
        return construction

    def showConstruction(self, surface, t):
        """Show the construction of the form at the t position."""
        construction = self.getConstruction(t)
        l = len(construction)
        for i in range(l):
            k = 255 * (i + 1) / l
            color = (0, 0, k)
            for segment in construction[i]:
                # print(segment)
                segment.show(surface, color=color, width=2)

    def showPoints(self, surface, points, color=None):
        """Show the points on the surface."""
        if not color:
            color = self.point_color
        for point in points:
            point.show(surface, color=color, radius=0.01)

    def showSegments(self, surface, segments, color=None):
        """Show the segments on the surface."""
        if not color:
            color = self.segment_color
        for segment in segments:
            segment.show(surface, color=color, width=2)

    def getSegments(self):
        """Return the segments of the trajectory of the curve."""
        return [Segment(points[i], points[i + 1]) for i in range(len(self.points) - 1)]

    segments = property(getSegments)


class BezierForm(Form):
    def show(self, context):
        """Show the curved form on the screen."""
        self.bezier.show(context)

    @property
    def bezier(self):
        """Return the bezier curve of the form."""
        return BezierCurve(self.points + [self.points[0]])


class Arrow(BezierCurve):
    """Join 2 objects with an arrow."""

    def __init__(
        self,
        points,
        point_color=colors.WHITE,
        segment_color=colors.WHITE,
        vector_color=colors.WHITE,
    ):
        super().__init__(self, points, point_color, segment_color, vector_color)

    def show(self, surface, p=50):
        """Show the arrow on the screen."""
        points = [self(i / p) for i in range(p)]  # p=p+1 ?
        segments = [Segment(points[i], points[i + 1]) for i in range(p - 1)]
        vectors = [Vector(points[i], points[i + 1]) for i in range(0, p - 1, 5)]
        self.showPoints(surface, points)
        self.showSegments(surface, segments)
        self.showVectors(surface, vectors)

    def showVectors(self, surface, vectors, color=None):
        """Show the vectors on the surface."""
        if not color:
            color = self.vector_color
        for vector in vectors:
            vector.show(surface, color=color)


if __name__ == "__main__":
    from mycontext import Surface

    surface = Surface(name="Curves demonstration")
    l = 10
    points = [Point(2 * x, random.randint(-5, 5)) for x in range(l)]
    t = Trajectory(points, segment_color=colors.GREEN)
    b = BezierCurve(points, segment_color=colors.RED)
    # st=t.sampleSegments(3)
    # t1=Trajectory(st,segment_color=colors.BLUE)
    # print(b.points[-1])
    # print(b(1))
    # print(b.segments[-1])
    # c=CurvedForm(points)
    n = 0
    ncp = 50  # number construction points

    while surface.open:
        surface.check()
        surface.control()
        surface.clear()
        surface.show()

        Point.turnPoints([1 / 1000 for i in range(l)], points)
        n = (n + 1) % (ncp + 1)
        b.showConstruction(surface, n / ncp)

        t.show(surface)
        b.show(surface)

        p1 = b(n / ncp)
        p2 = Point(*t(n / ncp))

        # l1=Line.createFromTwoPoints(p1,p2)

        p1.show(surface, color=colors.YELLOW, radius=0.1, fill=True)
        p2.show(surface, color=colors.YELLOW, radius=0.1, fill=True)
        # l1.show(surface,color=colors.LIGHTGREY)

        surface.flip()



from math import pi, sqrt, cos, sin
from cmath import polar

from .tools import timer
from . import colors

import itertools
import random
import math
import copy

average = mean = lambda x: sum(x) / len(x)

digits = 2  # Number of digits of precision of the objects when displayed


class Point:
    """Representation of a point that can be displayed on screen."""

    @classmethod
    def sum(cls, points, **kwargs):
        """Return the points which components are the respectives sums of the
        components of the given points."""
        p = Point.null(**kwargs)
        for point in points:
            p += point
        return p

    @classmethod
    def average(cls, points):
        """Return the point which position is the average of
        the position of the given points"""
        return Point.sum(points) / len(points)

    @classmethod
    def origin(cls, d=2, **kwargs):
        """Return the origin."""
        return cls([0 for i in range(d)], **kwargs)

    null = neutral = zero = origin

    @classmethod
    def random(cls, d=2, borns=[-1, 1], **kwargs):
        """Create a random point using optional minimum and maximum."""
        components = [random.uniform(*borns) for i in range(d)]
        return cls(*components, **kwargs)

    @staticmethod
    def distance(p1, p2):
        """Return the distance between the two points."""
        return math.hypot(p1.x - p2.x, p1.y - p2.y)

    @staticmethod
    def turnPoints(angles, points):
        """Turn the points around themselves."""
        l = len(points)
        for i in range(l - 1):
            points[i].turn(angles[i], points[i + 1 :])

    @staticmethod
    def showPoints(surface, points):
        """Show the points on the surface."""
        for point in points:
            point.show(surface)

    @staticmethod
    def closest(point, points):
        """Return the point of the list of points that is the closest to the given point."""
        pts = [(Point.distance(point, p), p) for p in points]
        pts.sort()
        return pts[0][1]

    @staticmethod
    def farthest(point, points):
        """Return the point of the list of points that is the farthest to the given point."""
        # print(point)
        pts = [(Point.distance(point, p), p) for p in points]
        pts.sort()
        # print(" ".join(map(lambda pt: str(pt[0])[:5]+","+str(pt[1]), pts)))
        return pts[-1][1]

    @classmethod
    def createFromVector(cls, vector):
        """Create a point from a vector."""
        return cls(vector.x, vector.y)

    def __init__(
        self,
        *components,
        mode=0,
        size=[0.1, 0.1],
        width=1,
        radius=0.02,
        fill=False,
        color=colors.WHITE,
        conversion=True
    ):
        """Create a point using its components and optional radius, fill, color and conversion."""
        if components != ():
            if type(components[0]) == list:
                components = components[0]
        self.components = list(components)
        self.mode = mode
        self.size = size
        self.width = width
        self.radius = radius
        self.fill = fill
        self.color = color
        self.conversion = conversion

    def __hash__(self):
        """Return a single number using the hash of its attributes."""
        hash(self.components)

    def set(self, point):
        """Set the components of the point to the components of an other given point."""
        self.components = [c for c in point]

    def __len__(self):
        """Return the number of components of the point."""
        return len(self.components)

    def setX(self, value):
        """Set the x component."""
        self.components[0] = value

    def getX(self):
        """Return the x component."""
        return self.components[0]

    def delX(self):
        """Delete the x component and so shifting to a new one."""
        del self.components[0]

    def setY(self, value):
        """Set the y component."""
        self.components[1] = value

    def getY(self):
        """Return the y component."""
        return self.components[1]

    def delY(self):
        """Delete the y component."""
        del self.components[1]

    x = property(
        getX, setX, delX, "Allow the user to manipulate the x component easily."
    )
    y = property(
        getY, setY, delY, "Allow the user to manipulate the y component easily."
    )

    def __eq__(self, other):
        """Determine if two points are equals by comparing its components."""
        return abs(self - other) < 10e-10

    def __ne__(self, other):
        """Determine if two points are unequals by comparing its components."""
        return tuple(self) != tuple(other)

    def __rmul__(self, r):
        """Multiplication by a scalar like vectors."""
        return Point(*[r * c for c in self.components])

    def __truediv__(self, r):
        """Divide a points by a scalar."""
        return Point(*[c / r for c in self.components])

    def __contains__(self, other):
        """Determine if an object is in the point."""
        x, y = other
        return self.radius >= sqrt((x - self.x) ** 2 + (y - self.y) ** 2)

    def __getitem__(self, index):
        """Return x or y value using given index."""
        return self.components[index]

    def __setitem__(self, index, value):
        """Change x or y value using given index and value."""
        self.components[index] = value

    def __abs__(self):
        """Return the distance of the point to the origin."""
        return Vector.createFromPoint(self).norm

    def __tuple__(self):
        """Return the components in tuple form."""
        return tuple(self.components)

    def __list__(self):
        """Return the components."""
        return self.components

    def rotate(self, angle=pi, point=None):
        """Rotate the point using the angle and the center of rotation.
        Uses the origin for the center of rotation by default."""
        if not point:
            point = Point.origin(d=self.dimension)
        v = Vector.createFromTwoPoints(point, self)
        v.rotate(angle)
        self.components = v(point).components

    def turn(self, angle=pi, points=[]):
        """Turn the points around itself."""
        for point in points:
            point.rotate(angle, self)

    def move(self, *step):
        """Move the point using given step."""
        self.x += step[0]
        self.y += step[1]

    def around(self, point, distance):
        """Determine if a given point is in a radius 'distance' of the point."""
        return self.distance(point) <= distance

    def showCross(self, window, color=None, size=None, width=None, conversion=None):
        """Show the point under the form of a cross using the window."""
        if not color:
            color = self.color
        if not size:
            size = self.size
        if not width:
            width = self.width
        if not conversion:
            conversion = self.conversion
        x, y = self
        sx, sy = size
        xmin = x - sx / 2
        ymin = y - sy / 2
        xmax = x + sx / 2
        ymax = y + sy / 2
        window.draw.line(
            window.screen, color, [xmin, ymin], [xmax, ymax], width, conversion
        )
        window.draw.line(
            window.screen, color, [xmin, ymax], [xmax, ymin], width, conversion
        )

    def showCircle(self, window, color=None, radius=None, fill=None, conversion=None):
        """Show a point under the form of a circle using the window."""
        if not color:
            color = self.color
        if not radius:
            radius = self.radius
        if not fill:
            fill = self.fill
        if not conversion:
            conversion = self.conversion
        window.draw.circle(
            window.screen, color, [self.x, self.y], radius, fill, conversion
        )

    def show(
        self,
        window,
        color=None,
        mode=None,
        fill=None,
        radius=None,
        size=None,
        width=None,
        conversion=None,
    ):
        """Show the point on the window."""
        if not mode:
            mode = self.mode
        if mode == 0 or mode == "circle":
            self.showCircle(window, color, radius, fill, conversion)
        if mode == 1 or mode == "cross":
            self.showCross(window, color, size, width, conversion)

    def showText(self, context, text, size=1, color=colors.WHITE, conversion=True):
        """Show the text next to the point on the window."""
        context.print(text, self.components, size, color=color, conversion=conversion)

    def __add__(self, other):
        """Add two points."""
        return Point([c1 + c2 for (c1, c2) in zip(self, other)])

    def __iadd__(self, other):
        """Add a point to the actual point."""
        self.components = [c1 + c2 for (c1, c2) in zip(self, other)]
        return self

    __radd__ = __add__

    def __sub__(self, other):
        """Add a point to the actual point."""
        return Point([c1 - c2 for (c1, c2) in zip(self, other)])

    def __isub__(self, other):
        """Add a point to the actual point."""
        self.components = [c1 - c2 for (c1, c2) in zip(self, other)]
        return self

    __rsub__ = __sub__

    def __sub__(self, other):
        """Substract the components of 2 objects."""
        return Point(self.x - other[0], self.y - other[1])

    def __ge__(self, other):
        """Determine if a point is farther to the origin."""
        return self.x ** 2 + self.y ** 2 >= other.x ** 2 + other.y ** 2

    def __gt__(self, other):
        """Determine if a point is farther to the origin."""
        return self.x ** 2 + self.y ** 2 > other.x ** 2 + other.y ** 2

    def __le__(self, other):
        """Determine if a point is the nearest to the origin."""
        return self.x ** 2 + self.y ** 2 <= other.x ** 2 + other.y ** 2

    def __lt__(self, other):
        """Determine if a point is the nearest to the origin."""
        return self.x ** 2 + self.y ** 2 < other.x ** 2 + other.y ** 2

    def __iter__(self):
        """Iterate the points of the form."""
        self.iterator = 0
        return self

    def __next__(self):
        """Return the next point threw an iteration."""
        if self.iterator < len(self.components):
            self.iterator += 1
            return self.components[self.iterator - 1]
        else:
            raise StopIteration

    def truncate(self):
        """Truncate the position of the point by making the x and y components integers."""
        for i in range(self.dimension):
            self.components[i] = int(self.components[i])

    def __str__(self):
        """Return the string representation of a point."""
        return "p(" + ",".join([str(round(c, digits)) for c in self.components]) + ")"

    def getPosition(self):
        """Return the components."""
        return self.components

    def setPosition(self, position):
        """Set the components."""
        self.components = position

    def getDimension(self):
        """Return the dimension of the point."""
        return len(self.components)

    def setDimension(self, dimension):
        """Set the dimension of the point by setting to 0 the new components."""
        self.components = self.components[:dimension]
        self.components += [0 for i in range(dimension - len(self.components))]

    def delDimension(self):
        """Delete the components of the points."""
        self.components = []

    position = property(
        getPosition,
        setPosition,
        "Same as component although only component should be used.",
    )
    dimension = property(
        getDimension,
        setDimension,
        delDimension,
        "Representation of the dimension point which is the length of the components.",
    )


class Direction:
    """Base class of lines and segments."""

    pass


class Vector:
    @classmethod
    def null(cls, d=2):
        """Return the null vector."""
        return cls(*[0 for i in range(d)])

    neutral = zero = null

    @classmethod
    def random(cls, d=2, borns=[-1, 1], **kwargs):
        """Create a random vector using optional min and max."""
        components = [random.uniform(*borns) for i in range(d)]
        return cls(*components, **kwargs)

    @classmethod
    def sum(cls, vectors):
        """Return the vector that correspond to the sum of all the vectors."""
        result = cls.null()
        for vector in vectors:
            result += vector
        return result

    @classmethod
    def average(cls, vectors):
        """Return the vector that correspond to the mean of all the vectors."""
        return cls.sum(vectors) / len(vectors)

    mean = average

    @classmethod
    def collinear(cls, *vectors, e=10e-10):
        """Determine if all the vectors are colinear."""
        l = len(vectors)
        if l == 2:
            v1 = vectors[0]
            v2 = vectors[1]
            return abs(v1.x * v2.y - v1.y - v2.x) < e
        else:
            for i in range(l):
                for j in range(i + 1, l):
                    if not cls.collinear(vectors[i], vectors[j]):
                        return False
            return True

    @classmethod
    def sameDirection(cls, *vectors, e=10e-10):
        """Determine if all the vectors are in the same direction."""
        l = len(vectors)
        if l == 2:
            v1 = vectors[0]
            v2 = vectors[1]
            return (abs(v1.angle - v2.angle) % (2 * math.pi)) < e
        else:
            for i in range(l):
                for j in range(i + 1, l):
                    if not cls.sameDirection(vectors[i], vectors[j]):
                        return False
            return True

    @classmethod
    def createFromPolar(cls, norm, angle, **kwargs):
        """Create a vector using norm and angle from polar coordonnates."""
        x, y = cls.cartesian([norm, angle])
        return cls(x, y, **kwargs)

    @classmethod
    def createFromSegment(cls, segment, **kwargs):
        """Create a vector from a segment."""
        return cls.createFromTwoPoints(segment.p1, segment.p2, **kwargs)

    @classmethod
    def createFromTwoPoints(cls, point1, point2, **kwargs):
        """Create a vector from 2 points."""
        return cls([c2 - c1 for (c1, c2) in zip(point1, point2)], **kwargs)

    @classmethod
    def createFromTwoTuples(cls, tuple1, tuple2, **kwargs):
        """Create a vector from 2 tuples."""
        return cls([c2 - c1 for (c1, c2) in zip(tuple1, tuple2)], **kwargs)

    @classmethod
    def createFromPoint(cls, point, **kwargs):
        """Create a vector from a single point."""
        return cls(point.x, point.y, **kwargs)

    @classmethod
    def createFromLine(cls, line, **kwargs):
        """Create a vector from a line."""
        angle = line.angle
        x, y = cls.cartesian([1, angle])
        return cls(x, y, **kwargs)

    @staticmethod
    def polar(position):
        """Return the polar position [norm,angle] using cartesian position [x,y]."""
        return list(polar(complex(position[0], position[1])))

    @staticmethod
    def cartesian(position):
        """Return the cartesian position [x,y] using polar position [norm,angle]."""
        return [position[0] * cos(position[1]), position[0] * sin(position[1])]

    def __init__(self, *components, color=colors.WHITE, width=1, arrow=[0.1, 0.5]):
        """Create a vector."""
        if components:
            if isinstance(components[0], list):
                components = components[0]
        self.components = list(components)
        self.color = color
        self.width = width
        self.arrow = arrow

    def set(self, v):
        """Set a vector to the values of another without changing its color, with or arrow."""
        self.components = v.components

    def setNull(self):
        """Set the components of the vector to zero."""
        self.components = [0 for i in range(len(self.components))]

    # X component
    def setX(self, value):
        """Set the x component."""
        self.components[0] = value

    def getX(self):
        """Return the x component."""
        return self.components[0]

    def delX(self):
        """Delete the x component and so shifting to a new one."""
        del self.components[0]

    # Y component
    def setY(self, value):
        """Set the y component."""
        self.components[1] = value

    def getY(self):
        """Return the y component."""
        return self.components[1]

    def delY(self):
        """Delete the y component."""
        del self.components[1]

    # Angle
    def getAngle(self):
        """Return the angle of a vector in polar coordinates."""
        x, y = self.components
        return math.atan2(y, x)

    def setAngle(self, value):
        """Change the angle of the points without changing its norm."""
        n, a = Vector.polar(self.components)
        self.components = Vector.cartesian([n, value])

    def delAngle(self):
        """Set to zero the angle of the vector."""
        self.setAngle(0)

    # Norm
    def getNorm(self, norm=lambda l: (sum(map(lambda x: x ** 2, l))) ** (1 / 2)):
        """Return the euclidian norm of the vector by default."""
        return norm(self.components)

    def setNorm(self, value):
        """Change the angle of the points without changing its norm."""
        n, a = Vector.polar(self.components)
        self.components = Vector.cartesian([value, a])

    # Position
    def getPosition(self):
        """Return the components."""
        return self.components

    def setPosition(self, position):
        """Set the components."""
        self.components = position

    def delPosition(self):
        """Set the vector to the null vector."""
        self.components = [0 for i in range(len(self.components))]

    x = property(
        getX, setX, delX, doc="Allow the user to manipulate the x component easily."
    )
    y = property(
        getY, setY, delY, doc="Allow the user to manipulate the y component easily."
    )
    norm = property(
        getNorm,
        setNorm,
        doc="Allow the user to manipulate the norm of the vector easily.",
    )
    angle = property(
        getAngle,
        setAngle,
        doc="Allow the user to manipulate the angle of the vector easily.",
    )
    position = property(getPosition, setPosition, doc="Same as components.")

    def limit(self, n):
        """Limit the norm of the vector by n."""
        if self.norm > n:
            self.norm = n

    def show(self, context, p=Point.neutral(), color=None, width=None):
        """Show the vector."""
        if not color:
            color = self.color
        if not width:
            width = self.width
        q = self(p)
        v = -self * self.arrow[0]  # wtf
        v1 = v % self.arrow[1]
        v2 = v % -self.arrow[1]
        a = v1(q)
        b = v2(q)
        context.draw.line(context.screen, color, p.components, q.components, width)
        context.draw.line(context.screen, color, q.components, a.components, width)
        context.draw.line(context.screen, color, q.components, b.components, width)

    def showFromTuple(self, context, t=(0, 0), **kwargs):
        """Show a vector from a tuple."""
        p = Point(*t)
        self.show(context, p, **kwargs)

    def showText(self, surface, point, text, color=None, size=20):
        """Show the text next to the vector."""
        if not color:
            color = self.color
        v = self / 2
        point = v(point)
        surface.print(text, tuple(point), color=color, size=size)

    def __len__(self):
        """Return the number of components."""
        return len(self.components)

    def __iter__(self):
        """Iterate the points of the form."""
        self.iterator = 0
        return self

    def __next__(self):
        """Return the next point threw an iteration."""
        if self.iterator < len(self.components):
            self.iterator += 1
            return self.components[self.iterator - 1]
        else:
            raise StopIteration

    def __neg__(self):
        """Return the negative vector."""
        return Vector([-c for c in self.components])

    def colinear(self, other, e=10e-10):
        """Return if two vectors are colinear."""
        return abs(self.x * other.y - self.y * other.x) < e

    __floordiv__ = colinear

    def scalar(self, other):
        """Return the scalar product between two vectors."""
        return self.x * other.x + self.y * other.y

    def cross(self, other):
        """Determine if a vector crosses another using dot product."""
        return self.scalar(other) == 0

    def __mul__(self, factor):
        """Multiply a vector by a given factor."""
        if type(factor) == int or type(factor) == float:
            return Vector([c * factor for c in self.components])
        else:
            raise NotImplementedError
            raise Exception(
                "Type "
                + str(type(factor))
                + " is not valid. Expected float or int types."
            )

    __imul__ = (
        __rmul__
    ) = __mul__  # Allow front extern multiplication using back extern multiplication with scalars

    def __truediv__(self, factor):
        """Multiply a vector by a given factor."""
        if type(factor) == Vector:
            raise NotImplementedError
        else:
            return Vector([c / factor for c in self.components])

    def __itruediv__(self, factor):
        """Divide the components of a vector by a given factor."""
        self.components = [c / factor for c in self.components]
        return self

    def __add__(self, other):
        """Add two vector together."""
        return Vector([c1 + c2 for (c1, c2) in zip(self.components, other.components)])

    def __sub__(self, other):
        """Sub two vector together."""
        return Vector([c1 - c2 for (c1, c2) in zip(self.components, other.components)])

    def __iadd__(self, other):
        """Add a vector to another."""
        self.components = [
            c1 + c2 for (c1, c2) in zip(self.components, other.components)
        ]
        return self

    def __isub__(self, other):
        """Substract a vector to another."""
        self.components = [
            c1 - c2 for (c1, c2) in zip(self.components, other.components)
        ]
        return self

    __radd__ = __add__
    __rsub__ = __sub__

    def rotate(self, angle):
        """Rotate a vector using the angle of rotation."""
        n, a = Vector.polar([self.x, self.y])
        a += angle
        self.x = n * cos(a)
        self.y = n * sin(a)

    def __mod__(self, angle):
        """Return the rotated vector using the angle of rotation."""
        n, a = Vector.polar([self.x, self.y])
        a += angle
        return Vector(
            n * cos(a), n * sin(a), color=self.color, width=self.width, arrow=self.arrow
        )

    __imod__ = __mod__

    def __getitem__(self, index):
        """Return x or y value using given index."""
        return self.position[index]

    def __setitem__(self, index, value):
        """Change x or y value using given index and value."""
        self.position[index] = value

    def __call__(self, *points):
        """Return points by applying the vector on those."""
        if points != ():
            if type(points[0]) == list:
                points = points[0]
        if len(points) == 0:
            raise Exception(
                "A vector can only be applied to a point or a list of points."
            )
        elif len(points) == 1:
            return points[0] + self
        else:
            return [point + self for point in points]

    def applyToPoint(self, point):
        """Return the point after applying the vector to it."""
        return self + point

    def applyToPoints(self, points):
        """Return the points after applying the vector to those."""
        return [point + self for point in points]

    def __xor__(self, other):
        """Return the angle between two vectors."""
        return self.angle - other.angle

    def __invert__(self):
        """Return the unit vector."""
        a = self.angle
        x, y = Vector.cartesian([1, a])
        return Vector(x, y)

    unit = property(__invert__)

    def __str__(self):
        """Return a string description of the vector."""
        return "v(" + ",".join([str(round(c, digits)) for c in self.components]) + ")"

    def __tuple__(self):
        """Return the components in tuple form."""
        return tuple(self.components)

    def __list__(self):
        """Return the components."""
        return self.components


class Segment(Direction):
    @classmethod
    def null(cls):
        """Return the segment whoose points are both the origin."""
        return cls(*[Point.origin() for i in range(2)])

    @classmethod
    def random(cls, d=2, borns=[-1, 1], **kwargs):
        """Create a random segment."""
        p1 = Point.random(d, borns)
        p2 = Point.random(d, borns)
        return cls(*[p1, p2], **kwargs)

    @classmethod
    def createFromTuples(cls, *tps, **kwargs):
        """Create a segment using tuples and optional arguments."""
        pts = [Point(*tp) for tp in tps]
        return cls(*pts, **kwargs)

    def __init__(self, *points, width=1, color=colors.WHITE, conversion=True):
        """Create the segment using 2 points, width and color."""
        if (
            len(points) > 0
        ):  # Extracting the points arguments under the same list format
            if type(points[0]) == list:
                points = points[0]
        if len(points) == 1:
            points = points[0]
        if len(points) != 2:
            raise Exception("A segment must have 2 points.")
        self.points = list(points)
        self.width = width
        self.color = color
        self.conversion = conversion

    def __str__(self):
        """Return the string representation of a segment."""
        return "s(" + str(self.p1) + "," + str(self.p2) + ")"

    def __call__(self, t=1 / 2):
        """Return the point C of the segment so that Segment(p1,C)=t*Segment(p1,p2)."""
        return (t * self.vector)(self.p1)

    def sample(self, n, include=True):
        """Sample n points of the segment.
        It is also possible to include the last point if wanted."""
        return [self(t / n) for t in range(n + int(include))]

    __rmul__ = __imul__ = __mul__ = lambda self, t: Segment(self.p1, self(t))

    def getCenter(self):
        """Return the center of the segment in the general case."""
        return Point.average(self.points)

    def setCenter(self, np):
        """Set the center of the segment."""
        p = self.getCenter()
        v = Vector.createFromTwoPoints(p, np)
        for i in range(len(self.points)):
            self.points[i] = v(self.points[i])

    def getAngle(self):
        """Return the angle of the segment."""
        return self.vector.angle

    def setAngle(self, angle):
        """Set the angle of the segment."""
        self.vector.angle = angle

    def show(self, context, color=None, width=None, conversion=None):
        """Show the segment using window."""
        if color is None:
            color = self.color
        if width is None:
            width = self.width
        if conversion is None:
            conversion = self.conversion
        context.draw.line(
            context.screen, color, self.p1, self.p2, width, conversion=conversion
        )

    def showInBorders(self, window, color=None, width=None):
        """Show the segment within the boundaries of the window."""
        # It it really slow and it doesn't work as expected.
        xmin, ymin, xmax, ymax = window.getCorners()
        p = [Point(xmin, ymin), Point(xmax, ymin), Point(xmax, ymax), Point(xmin, ymax)]
        f = Form(p)
        if (self.p1 in f) and (self.p2 in f):
            window.draw.line(
                window.screen,
                color,
                [self.p1.x, self.p1.y],
                [self.p2.x, self.p2.y],
                width,
            )
        elif (self.p1 in f) and not (self.p2 in f):
            v = Vector.createFromTwoPoints(self.p1, self.p2)
            hl = HalfLine(self.p1, v.angle)
            p = f.crossHalfLine(hl)
            if p:
                print(len(p))
                p = p[0]
                window.draw.line(
                    window.screen, color, [self.p1.x, self.p1.y], [p.x, p.y], width
                )
        elif not (self.p1 in f) and (self.p2 in f):
            v = Vector.createFromTwoPoints(self.p2, self.p1)
            hl = HalfLine(self.p2, v.angle)
            p = f.crossHalfLine(hl)
            if p:
                print(len(p))
                p = p[0]
                window.draw.line(
                    window.screen, color, [p.x, p.y], [self.p2.x, self.p2.y], width
                )
        else:
            ps = f.crossSegment(self)
            if len(ps) == 2:
                p1, p2 = ps
                window.draw.line(
                    window.screen, color, [p1.x, p1.y], [p2.x, p2.y], width
                )

    def __contains__(self, point, e=1e-10):
        """Determine if a point is in a segment."""
        if point == self.p1:
            return True
        v1 = Vector.createFromTwoPoints(self.p1, point)
        v2 = self.getVector()
        return (abs(v1.angle - v2.angle) % (2 * math.pi) < e) and (v1.norm <= v2.norm)

    def __len__(self):
        """Return the number of points."""
        return len(self.points)

    def __iter__(self):
        """Iterate the points of the form."""
        self.iterator = 0
        return self

    def __next__(self):
        """Return the next point through an iteration."""
        if self.iterator < len(self.points):
            self.iterator += 1
            return self.points[self.iterator - 1]
        else:
            raise StopIteration

    def __getitem__(self, index):
        """Return the point corresponding to the index given."""
        return [self.points][index]

    def __setitem__(self, index, value):
        """Change the value the point corresponding value and index given."""
        self.points[index] = value

    def getLine(self, **kwargs):
        """Return the line through the end points of the segment."""
        return Line(self.p1, self.angle, **kwargs)

    def getVector(self):
        """Return the vector that goes from p1 to p2."""
        return Vector.createFromTwoPoints(self.p1, self.p2)

    def setVector(self, vector):
        """Set the vector that goes from p1 to p2."""
        self.p2 = vector(self.p1)

    def getLength(self):
        """Return the length of the segment."""
        return self.vector.norm

    def setLength(self, length):
        """Set the length of the segment."""
        self.vector.norm = length

    def rotate(self, angle, point=None):
        """Rotate the segment using an angle and an optional point of rotation."""
        if not point:
            point = self.middle
        self.p1.rotate(angle, point)
        self.p2.rotate(angle, point)

    def __or__(self, other):
        """Return bool for (2 segments are crossing)."""
        if isinstance(other, Segment):
            return self.crossSegment(other)
        elif isinstance(other, Line):
            return self.crossLine(other)
        elif isinstance(other, HalfLine):
            return other.crossSegment(self)
        elif isinstance(other, Form):
            return other.crossSegment(self)
        else:
            raise TypeError(
                "The collisions {}|{} are not dealt with.".format(
                    type(self), type(other)
                )
            )

    def getXmin(self):
        """Return the minimum of x components of the 2 end points."""
        return min(self.p1.x, self.p2.x)

    def getYmin(self):
        """Return the minimum of y components of the 2 ends points."""
        return min(self.p1.y, self.p2.y)

    def getXmax(self):
        """Return the maximum of x components of the 2 end points."""
        return max(self.p1.x, self.p2.x)

    def getYmax(self):
        """Return the maximum of y components of the 2 end points."""
        return max(self.p1.y, self.p2.y)

    def getMinimum(self):
        """Return the lower point"""
        v1 = Vector(*self.p1)
        v2 = Vector(*self.p2)
        if v1.angle < v2.angle:
            return self.p1
        else:
            return self.p2

    def getMaximum(self):
        """Return the upper point."""
        v1 = Vector(*self.p1)
        v2 = Vector(*self.p2)
        if v1.angle >= v2.angle:
            return self.p1
        else:
            return self.p2

    def getCorners(self):
        """Return the minimum and maximum of x and y components of the 2 end points."""
        return self.minimum.components + self.maximum.components

    def parallel(self, other):
        """Determine if the line is parallel to another object (line or segment)."""
        return other.angle == self.angle

    def cross(self, other, **kwargs):
        if isinstance(other, Segment):
            return self.crossSegment(other, **kwargs)
        elif isinstance(other, Line):
            return other.crossSegment(self, **kwargs)
        else:
            raise TypeError(
                "The collisions {}|{} are not dealt with.".format(
                    type(self), type(other)
                )
            )

    def crossSegment(self, other, e=1e-14):
        """Return the intersection point of the segment with another segment."""
        sl = self.getLine()
        ol = other.getLine()
        point = sl.crossLine(ol)
        if point is not None:
            if self.__contains__(point, e) and other.__contains__(point, e):
                return point

    def crossLine(self, other):
        """Return the intersection point of the segment with a line."""
        if self.parallel(other):
            return None
        line = self.getLine()
        point = other.crossLine(line)
        if point is not None:
            if point in self and point in other:
                return point

    def getP1(self):
        """Return the first point of the segment."""
        return self.points[0]

    def setP1(self, p1):
        """Set the first point of the segment."""
        self.points[0] = p1

    def getP2(self):
        """Return the second point of the segment."""
        return self.points[1]

    def setP2(self, p2):
        """Set the second point of the segment."""
        self.points[1] = p2

    p1 = property(getP1, setP1, doc="First point of the segment.")
    p2 = property(getP2, setP2, doc="Second point of the segment.")
    middle = center = property(getCenter, setCenter, doc="Center of the segment.")
    vector = property(getVector, setVector, doc="Unit vector of the segment.")
    angle = property(getAngle, setAngle, doc="Angle of the segment.")
    length = property(getLength, setLength, doc="Length of the segment.")
    line = property(getLine, doc="Line passing by both extremities.")
    xmin = property(getXmin, doc="xmin")
    ymin = property(getYmin, doc="ymin")
    xmax = property(getXmax, doc="xmax")
    ymax = property(getYmax, doc="ymax")
    minimum = property(getMinimum, doc="Point at the right")
    maximum = property(getMaximum, doc="Point at the left")


class Line(Direction):
    @classmethod
    def random(cls, borns=[-1, 1], angle_borns=[-math.pi, math.pi], **kwargs):
        """Return a random line."""
        point = Point.random(borns=borns)
        angle = random.uniform(*angle_borns)
        return cls(point, angle, **kwargs)

    @classmethod
    def createFromPointAndVector(cls, point, vector, **kwargs):
        """Create a line using a point and a vector with optional features."""
        return cls(point, vector.angle, **kwargs)

    @classmethod
    def createFromTwoPoints(cls, point1, point2, **kwargs):
        """Create a line using two points with optional features."""
        vector = Vector.createFromTwoPoints(point1, point2)
        return cls(point1, vector.angle, **kwargs)

    def __init__(self, point, angle, width=1, color=colors.WHITE, correct=True):
        """Create the line using a point and a vector with optional width and color.
        The line is uses a unique system of components [neighbour point, angle].
        The neighbour point is the nearest point to (0,0) that belongs to the line.
        The angle is the orientated angle between the line itself and another line parallel
        to the x-axis and crossing the neighbour point. Its range is [-pi/2,pi/2[ which makes it unique."""
        if correct:
            self.angle = angle
            self.point = point
        else:
            self._angle = angle
            self._point = point
        self.width = width
        self.color = color

    def __str__(self, precision=2):
        """Return a string representation of the line."""
        return (
            "l(a="
            + str(round(self.slope, precision))
            + ",b="
            + str(round(self.ordinate, precision))
            + ")"
        )

    def __call__(self, x):
        """Evaluate the line like a linear function in cartesian coordinates."""
        return self.slope * x + self.ordinate

    def __eq__(self, l):
        """Determine if two lines are the same."""
        return l.point == self.point and l.angle == self.angle

    def getCompleteCartesianCoordonnates(self):
        """Return a,b,c according to the cartesian equation of the line: ax+by+c=0."""
        v = self.vector
        p1 = self.point
        p2 = v(self.point)
        if v.x == 0:
            a = 1
            b = 0
            c = -p1.x
        else:
            a = -(p1.y - p2.y) / (p1.x - p2.x)
            b = 1
            c = -(a * p1.x + b * p1.y)
        return a, b, c

    def getReducedCartesianCoordonnates(self):
        """Return a,b according to the reduced cartesian equation of the line: y=ax+b."""
        return self.slope, self.ordinate

    def getAngle(self):
        """Return the angle of the line."""
        return self._angle

    def setAngle(self, angle):
        """Set the angle of the line."""
        self._angle = (angle + math.pi / 2) % math.pi - math.pi / 2
        # self._angle = angle % (2*math.pi)

    angle = property(
        getAngle,
        setAngle,
        doc="Representation of the angle of the line after correction.",
    )

    def rotate(self, angle, point=Point(0, 0)):
        """Rotate the line."""
        self.angle += angle  # Incomplete

    def getPoint(self):
        """Return the neighbour point."""
        return self._point

    def setPoint(self, point):
        """Set the neighbour point to another one."""
        self._point = point
        self._point = self.projectPoint(Point.origin(point.dimension))

    point = property(getPoint, setPoint, doc="Neighbour point of the line.")

    def getUnitVector(self):
        """Return the unit vector of the line."""
        return Vector.createFromPolar(1, self.angle)

    def setUnitVector(self, vector):
        """Set the unit vector of the line."""
        self.angle = vector.angle

    def getNormalVector(self):
        """Return the normal vector of the line."""
        vector = self.unit_vector
        vector.rotate(math.pi / 2)
        return vector

    def setNormalVector(self, vector):
        """Set the normal vector of the line."""
        self.angle = vector.angle + math.pi / 2

    def getSlope(self):
        """Return the slope of the line."""
        return math.tan(self.angle)

    def setSlope(self, slope):
        """Set the slope of the line by changing its angle and point."""
        self.angle = math.atan(slope)

    def getOrdinate(self):
        """Return the ordinate of the line."""
        return self.point.y - self.slope * self.point.x

    def setOrdinate(self, ordinate):
        """Set the ordinate of the line by changing its position."""
        if abs(self.angle) == math.pi / 2:
            raise ValueError(
                "Impossible to set an ordinate because the line is parallel to the y axis. "
            )
        self.point.y += ordinate

    def getFunction(self):
        """Return the affine function that correspond to the line."""
        return lambda x: self.slope * x + self.ordinate

    def setFunction(self, function):
        """Set the function of the line by changing its slope and ordinate."""
        self.ordinate = function(0)
        self.slope = function(1) - function(0)

    def getReciproque(self):
        """Return the reciproque of the affine function that correspond to the line."""
        return lambda y: (y - self.ordinate) / self.slope

    def setReciproque(self, reciproque):
        """Return the reciproque function of the affine function."""
        x0, x1 = reciproque(0), reciproque(1)
        self.slope = 1 / (x1 - x0)
        self.ordinate = -x0 * self.slope

    def evaluate(self, x):
        """Evaluate the line as a affine function."""
        return self.function(x)

    def devaluate(self, y):
        """Evaluate the reciproque function of the affine funtion of the line."""
        return self.reciproque(y)

    def cross(self, other):
        """Return the point of intersection between the line and another object."""
        if isinstance(other, Line):
            return self.crossLine(other)
        elif isinstance(other, Segment):
            return self.crossSegment(other)
        elif isinstance(other, HalfLine):
            return other.crossLine(self)
        elif isinstance(other, Form):
            return other.crossLine(self)
        else:
            raise TypeError(
                "The collisions {}|{} are not dealt with.".format(
                    type(self), type(other)
                )
            )

    __or__ = cross

    def crossSegment(self, other, e=1e-14, **kwargs):
        """Return the point of intersection between a segment and the line."""
        # Determine the point of intersection between the line of the given segment ang the line
        line = other.getLine()
        point = self.crossLine(line)
        if point is None:
            return None
        x, y = point
        # Determine if the point of intersection belongs to both the segment and the line
        if (
            other.xmin - e <= point.x <= other.xmax + e
            and other.ymin - e <= y <= other.ymax + e
        ):
            return Point(x, y, **kwargs)
        # By default if nothing is returned the function returns None

    def crossLine(self, other):
        """Return the point of intersection between two lines with vectors
        calculation. Works in all cases even with vertical lines."""
        a, b = self.point
        c, d = other.point
        m, n = self.vector
        o, p = other.vector
        if n * o == m * p:  # The lines are parallels
            return None
        elif self.angle == -math.pi / 2:
            return Point(a, d)
        elif other.angle == -math.pi / 2:
            return Point(b, c)
        else:
            x = (a * n * o - b * m * o - c * m * p + d * m * o) / (n * o - m * p)
            y = (x - a) * n / m + b
            return Point(x, y)

    def parallel(self, other):
        """Determine if the line is parallel to another object (line or segment)."""
        return other.angle == self.angle

    def __contains__(self, point, e=10e-10):
        """Determine if a point belongs to the line."""
        v1 = self.vector
        v2 = Vector.createFromTwoPoints(self.point, point)
        return v1.colinear(v2, e)

    def getHeight(self, point):
        """Return the height line between the line and a point."""
        return Line(point, self.normal_vector.angle)

    def distanceFromPoint(self, point):
        """Return the distance between a point and the line."""
        return Vector.createFromTwoPoints(
            point, self.crossLine(self.getHeight(point))
        ).norm

    def projectPoint(self, point):
        """Return the projection of the point on the line."""
        vector = self.normal_vector
        angle = vector.angle
        line = Line(point, angle, correct=False)
        projection = self.crossLine(line)
        return projection

    def projectPoints(self, points):
        """Return the projections of the points on the line."""
        return [self.projectPoint(point) for point in points]

    def projectSegment(self, segment):
        """Return the projection of a segment on the line."""
        points = self.projectPoints(segment.points)
        return Segment(*points, segment.width, segment.color)

    def getSegmentWithinCorners(self, corners):
        """Return the segment made of the points of the line which are in the area
        delimited by the corners."""
        xmin, ymin, xmax, ymax = corners
        p1 = Point(xmin, ymin)
        p2 = Point(xmax, ymin)
        p3 = Point(xmax, ymax)
        p4 = Point(xmin, ymax)
        s1 = Segment(p1, p2)
        s2 = Segment(p2, p3)
        s3 = Segment(p3, p4)
        s4 = Segment(p4, p1)
        pts = []
        for s in [s1, s2, s3, s4]:
            cross = self.crossSegment(s)
            if cross:
                pts.append(cross)
        if len(pts) == 2:
            return Segment(*pts)

    def getPointsWithinCorners(self, corners):
        """Return the segment made of the points of the line which are in the area
        delimited by the corners."""
        xmin, ymin, xmax, ymax = corners
        p1 = Point(xmin, ymin)
        p2 = Point(xmax, ymin)
        p3 = Point(xmax, ymax)
        p4 = Point(xmin, ymax)
        v1 = Vector(p1, p2)
        v2 = Vector(p2, p3)
        v3 = Vector(p3, p4)
        v4 = Vector(p4, p1)
        l1 = Line.createFromPointAndVector(p1, v1)
        l2 = Line.createFromPointAndVector(p2, v2)
        l3 = Line.createFromPointAndVector(p3, v3)
        l4 = Line.createFromPointAndVector(p4, v4)
        lines = [l1, l3]
        points = []
        for line in lines:
            cross = self.crossLine(line)
            if cross:
                points.append(cross)
        if not points:
            lines = [l2, l4]
            for line in lines:
                cross = self.crossLine(line)
                if cross:
                    points.append(cross)
        return points

    def show(self, context, width=None, color=None):
        """Show the line on the context by drawing a segment greater than
        the size of the context."""
        if not color:
            color = self.color
        if not width:
            width = self.width
        s = max(context.size)
        u = min(context.units)
        n = s / u
        p = Point(*context.point())
        p = self.projectPoint(p)
        v1 = Vector.createFromPolar(n, self.angle)
        v2 = -v1
        s = Segment(v1(p), v2(p), color=color, width=width, conversion=True)
        s.show(context)

    def showWithinCorners(self, context, width=None, color=None):
        if not color:
            color = self.color
        if not width:
            width = self.width
        s = self.getSegmentWithinCorners(context.corners)
        s.show(context)

    vector = unit_vector = property(
        getUnitVector, setUnitVector, doc="Unit vector of the line."
    )
    normal_vector = property(
        getNormalVector, setNormalVector, doc="Normal vector of the line."
    )
    slope = property(getSlope, setSlope, doc="Slope of the line in cartesian system.")
    ordinate = property(
        getOrdinate, setOrdinate, doc="Ordinate of the line in cartesian system."
    )
    function = property(
        getFunction, setFunction, doc="Linear function of the line in cartesian system."
    )
    reciproque = property(
        getReciproque,
        setReciproque,
        doc="Reciproque of the function of the line in cartesian system.",
    )


class HalfLine(Line):
    def createFromLine(line):
        """Create a half line."""
        return HalfLine(line.point, line.angle)

    def __init__(self, point, angle, color=colors.WHITE, width=1):
        """Create a half line."""
        super().__init__(point, angle, color=color, width=width, correct=False)

    def getLine(self, correct=True):
        """Return the line that correspond to the half line."""
        return Line(self.point, self.angle, correct=correct)

    def getPoint(self):
        """Return the point of the half line."""
        return self.point

    def setPoint(self, point):
        """Set the point of the half line."""
        self.point = point

    def show(self, context, width=None, color=None):
        """Show the line on the surface."""
        if not color:
            color = self.color
        if not width:
            width = self.width
        xmin, ymin, xmax, ymax = context.corners
        form = Form.createFromTuples(
            [(xmin, ymin), (xmax, ymin), (xmax, ymax), (xmin, ymax)]
        )
        points = form.crossHalfLine(self)
        points += [self.point] * (2 - len(points))
        if len(points) > 0:
            context.draw.line(context.screen, color, points[0], points[1], width)

    def __contains__(self, point, e=10e-10):
        """Determine if a point is in the half line."""
        v1 = self.vector
        v2 = Vector.createFromTwoPoints(self.point, point)
        return abs(v1.angle - v2.angle) < e

    def cross(self, other):
        """Return the intersection point between the half line and another object."""
        if isinstance(other, Line):
            return self.crossLine(other)
        if isinstance(other, Segment):
            return self.crossSegment(other)
        if isinstance(other, HalfLine):
            return self.crossHalfLine(other)
        if isinstance(other, Form):
            return other.crossHalfLine(self)

    __or__ = cross

    def crossHalfLine(self, other):
        """Return the point of intersection of the half line with another."""
        ml = self.getLine(correct=False)
        ol = other.getLine(correct=False)
        point = ml.crossLine(ol)
        if point:
            if (point in self) and (point in other):
                return point

    def crossLine(self, other):
        """Return the point of intersection of the half line with a line."""
        ml = self.getLine(correct=False)
        point = ml.crossLine(other)
        if point:
            if (point in self) and (point in other):
                return point

    def crossSegment(self, other):
        """Return the point of intersection of the half line with a segment."""
        ml = self.getLine(correct=False)
        ol = other.getLine(correct=False)
        point = ml.crossLine(ol)
        if point:
            if (point in self) and (point in other):
                return point

    def __str__(self):
        """Return the string representation of a half line."""
        return "hl(" + str(self.point) + "," + str(self.angle) + ")"


class Form:
    @classmethod
    def random(cls, n=random.randint(5, 10), d=2, borns=[-1, 1], **kwargs):
        """Create a random form using the number of points 'n', the dimension
        of the points 'p' and their borns with optional arguments."""
        points = [Point.random(d=d, borns=borns) for i in range(n)]
        form = cls(points, **kwargs)
        form.makeSparse()
        return form

    def anyCrossing(forms):
        """Determine if any of the forms are crossing."""
        if len(forms) == 1:
            forms = forms[0]
        l = len(forms)
        for i in range(l):
            for j in range(i + 1, l):
                if forms[i].crossForm(forms[j]):
                    return True
        return False

    def allCrossing(forms):
        """Determine if all the forms are crossing."""
        if len(forms) == 1:
            forms = forms[0]
        l = len(forms)
        for i in range(l):
            for j in range(i + 1, l):
                if not forms[i].crossForm(forms[j]):
                    return False
        return True

    def cross(form1, form2):
        """Return the points of intersection between the crossing forms."""
        for point in form1.points:
            if point in form2:
                return True
        for point in form2.points:
            if point in form1:
                return True
        return

    @classmethod
    def intersectionTwoForms(cls, form1, form2):
        """Return the form which is the intersection of two forms."""
        if form1 == None:
            return form2
        if form2 == None:
            return form1
        if form1 == form2 == None:
            return None
        points = form1.crossForm(form2)
        if not points:
            return None
        for point in form1.points:
            if point in form2:
                points.append(point)
        for point in form2.points:
            if point in form1:
                points.append(point)
        form = cls(points)
        form.makeSparse()
        return form

    @classmethod
    def intersection(cls, forms):
        """Return the form which is the intersection of all the forms."""
        result = forms[0]
        for form in forms[1:]:
            result = cls.intersectionTwoForms(result, form)
        return result

    @classmethod
    def unionTwoForms(cls, form1, form2):
        """Return the union of two forms."""
        intersection_points = set(form1.crossForm(form2))
        if intersection_points:
            all_points = set(form1.points + form2.points)
            points = all_points.intersection(intersection_points)
            return [cls(points)]
        else:
            return [form1, form2]

    @classmethod
    def union(cls, forms):
        """Return the union of all forms."""
        """This function must be recursive."""
        if len(forms) == 2:
            return cls.unionTwoForms(forms[0], forms[1])
        else:
            pass

        result = forms[0]
        for form in forms[1:]:
            result.extend(cls.union(form, result))
        return result

    @classmethod
    def createFromTuples(cls, tps, conversion=True, radius=0.01, **kwargs):
        """Create a form from the tuples 'tps' and some optional arguments."""
        pts = [Point(*t, conversion=conversion, radius=0.01) for t in tps]
        return cls(pts, **kwargs)

    def __init__(
        self,
        points,
        fill=False,
        point_mode=0,
        point_size=[0.01, 0.01],
        point_radius=0.01,
        point_width=1,
        point_fill=False,
        side_width=1,
        color=None,
        point_color=colors.WHITE,
        side_color=colors.WHITE,
        area_color=colors.WHITE,
        cross_point_color=colors.WHITE,
        cross_point_radius=0.01,
        cross_point_mode=0,
        cross_point_width=1,
        cross_point_size=[0.1, 0.1],
        point_show=True,
        side_show=True,
        area_show=False,
    ):
        """Create the form object using points."""
        self.points = points

        self.point_mode = point_mode
        self.point_size = point_size
        self.point_width = point_width
        self.point_radius = point_radius
        self.point_color = point_color or color
        self.point_show = point_show
        self.point_fill = point_fill

        self.side_width = side_width
        self.side_color = side_color or color
        self.side_show = side_show

        self.area_color = area_color or color
        self.area_show = area_show or fill

        self.cross_point_color = cross_point_color
        self.cross_point_radius = cross_point_radius
        self.cross_point_mode = cross_point_mode
        self.cross_point_width = cross_point_width
        self.cross_point_size = cross_point_size

    def __str__(self):
        """Return the string representation of the form."""
        return "f(" + ",".join([str(p) for p in self.points]) + ")"

    def setFill(self, fill):
        """Set the form to fill its area when shown."""
        self.area_show = fill

    def getFill(self):
        """Return if the area is filled."""
        return self.area_show

    fill = property(
        getFill,
        setFill,
        doc="Allow the user to manipulate easily if the area is filled.",
    )

    def __iadd__(self, point):
        """Add a point to the form."""
        self.points.append(point)
        return self

    def __isub__(self, point):
        """Remove a point to the form."""
        self.points.remove(point)
        return self

    def __mul__(self, n):
        """Return a bigger form."""
        vectors = [n * Vector(*(p - self.center)) for p in self.points]
        return Form([vectors[i](self.points[i]) for i in range(len(self.points))])

    def __imul__(self, n):
        """Return a bigger form."""
        vectors = [n * Vector(*(p - self.center)) for p in self.points]
        self.points = [vectors[i](self.points[i]) for i in range(len(self.points))]
        return self

    __rmul__ = __mul__

    def __iter__(self):
        """Iterate the points of the form."""
        self.iterator = 0
        return self

    def __next__(self):
        """Return the next point threw an iteration."""
        if self.iterator < len(self.points):
            iterator = self.iterator
            self.iterator += 1
            return self.points[iterator]
        else:
            raise StopIteration

    def __eq__(self, other):
        """Determine if 2 forms are the same which check the equalities of their components."""
        return sorted(self.points) == sorted(other.points)

    def getCenter(self):
        return Point.average(self.points)

    def setCenter(self, center):
        """Set the center of the form."""
        p = center - self.center
        for i in range(len(self.points)):
            self.points[i] += p

    def getCentroid(self):
        """Return the point of the center.
        This only works for 2 dimensional forms obviously."""
        if len(self.points) == 0:
            # None
            return None
        elif len(self.points) == 1:
            # Same point
            return self.points[0]
        elif len(self.points) == 2:
            # Middle of a segment
            return Segment(*self.points).middle
        elif len(self.points) == 3:
            # Intersection point of 2 medians
            return Point.average(self.points)
        else:
            # Geometric decomposition to compute centroids (wikipedia)
            n = len(self.points)  # n is the number of points
            # There are n-2 forms
            forms = [
                Form([self.points[0]] + self.points[i : i + 2]) for i in range(1, n - 1)
            ]
            # So n-2 centroids and areas, except if some of the points are one upon another, no area is null
            centroids = [form.center for form in forms]
            areas = [form.area for form in forms]
            # we compute the average centroid weighted by the areas
            weighted_centroid = Point.sum([a * c for (c, a) in zip(centroids, areas)])
            centroid = weighted_centroid / sum(areas)
            return centroid

    def setCentroid(self, center):
        """Set the center of the form."""
        p = center - self.centroid
        for i in range(len(self.points)):
            self.points[i] += p

    def recenter(self, point=(0, 0)):
        """Recenter a form using the new center point."""
        self.center = Point(*point)

    def enlarge(self, n=2):
        """Enlarge the form by a factor of n."""
        c = self.center
        for i in range(len(self.points)):
            v = Vector.createFromTwoPoints(c, self.points[i])
            self.points[i].set((n * v)(c))

    def spread(self, n=2):
        """Take away the form by a factor of n."""
        for point in self.points:
            point *= n

    def getSegments(self):
        """"Return the list of the form sides."""
        l = len(self.points)
        return [
            Segment(
                self.points[i % l],
                self.points[(i + 1) % l],
                color=self.side_color,
                width=self.side_width,
            )
            for i in range(l)
        ]

    def setSegments(self, segments):
        """Set the segments of the form by setting its points to new values."""
        for point, segment in zip(self.points, segments):
            point.set(segment.p1)

    def getVectors(self):
        """Return the list of the form vectors."""
        l = len(self.points)
        return [
            Vector.createFromTwoPoints(
                self.points[i % l],
                self.points[(i + 1) % l],
                color=self.side_color,
                width=self.side_width,
            )
            for i in range(l)
        ]

    def setVectors(self, vectors):
        """Return the list of the form vectors."""
        l = len(self.points)
        for point, vector in zip(self.points, vectors):
            point.set(vector.components)

    def showAll(self, surface, **kwargs):
        """Show the form using a window."""
        # ,window,point_color=None,side_color=None,area_color=None,side_width=None,point_radius=None,color=None,fill=None,point_show=None,side_show=None
        if not "point_show" in kwargs:
            kwargs["point_show"] = self.point_show
        if not "side_show" in kwargs:
            kwargs["side_show"] = self.side_show
        if not "area_show" in kwargs:
            kwargs["area_show"] = self.area_show
        if kwargs["area_show"]:
            self.showAllArea(surface, **kwargs)
        if kwargs["side_show"]:
            self.showAllSegments(surface, **kwargs)
        if kwargs["point_show"]:
            self.showAllPoints(surface, **kwargs)

    def showFast(self, surface, point=None, segment=None, area=None):
        """Show the form using the surface and optional objects to show."""
        if point:
            self.showPoints(surface)
        if segment:
            self.showSegments(surface)
        if area:
            self.showArea(surface)

    def show(self, surface):
        """Show the form using the surface and optional objects to show."""
        if self.area_show:
            self.showArea(surface)
        if self.side_show:
            self.showSegments(surface)
        if self.point_show:
            self.showPoints(surface)

    def showFastArea(self, surface, color=None):
        """Show the area of the form using optional parameters such as the area
        of the color."""
        if not color:
            color = self.area_color
        ps = [tuple(p) for p in self.points]
        if len(ps) > 1:
            surface.draw.polygon(surface.screen, color, ps, False)

    def showAllArea(self, surface, **kwargs):
        """Show the area of the form using optional parameters such as the area
        of the color. This function is slower than the previous one because it
        checks if the dictionary or attributes contains the area_color."""
        if not "area_color" in kwargs:
            kwargs["area_color"] = self.area_color
        ps = [tuple(p) for p in self.points]
        if len(ps) > 1:
            surface.draw.polygon(surface.screen, kwargs["area_color"], ps, False)

    def showArea(self, surface):
        """Show the area of the form."""
        ps = [tuple(p) for p in self.points]
        if len(ps) > 1:
            surface.draw.polygon(surface.screen, self.area_color, ps, False)

    def showPoints(self, surface):
        """Show the points."""
        for point in self.points:
            point.show(surface)

    def showFastPoints(
        self,
        surface,
        color=None,
        mode=None,
        radius=None,
        size=None,
        width=None,
        fill=None,
    ):
        """Show the points of the form using optional parameters."""
        if color is None:
            color = self.point_color
        if radius is None:
            radius = self.point_radius
        if mode is None:
            mode = self.point_mode
        if size is None:
            size = self.point_size
        if width is None:
            width = self.point_width
        if fill is None:
            fill = self.point_fill
        for point in self.points:
            point.show(surface, color, mode, fill, radius, size, width)

    def showAllPoints(self, surface, **kwargs):
        """Show the points of the form using optional parameters.
        This method is slower than the previous one because it checks if the
        dictionary of attributes contains the arguments."""
        if not "point_color" in kwargs:
            kwargs["point_color"] = self.point_color
        if not "point_radius" in kwargs:
            kwargs["point_radius"] = self.point_radius
        if not "point_mode" in kwargs:
            kwargs["point_mode"] = self.point_mode
        if not "point_size" in kwargs:
            kwargs["point_size"] = self.point_size
        if not "point_width" in kwargs:
            kwargs["point_width"] = self.point_width
        if not "point_fill" in kwargs:
            kwargs["point_fill"] = self.point_fill
        for point in self.points:
            point.show(
                surface,
                color=kwargs["point_color"],
                mode=kwargs["point_mode"],
                fill=kwargs["point_fill"],
                radius=kwargs["point_radius"],
                size=kwargs["point_size"],
                width=kwargs["point_width"],
            )

    @timer
    def showFastSegments(self, context, color=None, width=None):
        """Show the segments of the form."""
        if not color:
            color = self.segment_color
        if not width:
            width = self.segment_width
        for segment in self.segments:
            segment.show(context, color, width)

    def showSegments(self, surface):
        """Show the segments without its parameters."""
        for segment in self.segments:
            segment.show(surface)

    def showAllSegments(self, surface, **kwargs):
        """Show the segments of the form."""
        if not "side_color" in kwargs:
            kwargs["side_color"] = self.side_color
        if not "side_width" in kwargs:
            kwargs["side_width"] = self.side_width
        for segment in self.segments:
            segment.show(
                surface, color=kwargs["side_color"], width=kwargs["side_width"]
            )

    def showFastCrossPoints(
        self, surface, color=None, mode=None, radius=None, width=None, size=None
    ):
        """Show the intersection points of the form crossing itself."""
        points = self.crossSelf()
        if not color:
            color = self.cross_point_color
        if not mode:
            mode = self.cross_point_mode
        if not radius:
            radius = self.cross_point_radius
        if not width:
            width = self.cross_point_width
        if not size:
            size = self.cross_point_size
        for point in points:
            point.show(
                surface, color=color, mode=mode, radius=radius, width=width, size=size
            )

    def showCrossPoints(self, surface):
        """Show the intersection points of the form crossing itself."""
        for point in self.cross_points:
            point.show(surface)

    def cross(self, other):
        """Return the points of intersections with the form and another object."""
        if isinstance(other, Form):
            return self.crossForm(other)
        elif isinstance(other, Segment):
            return self.crossSegment(other)
        elif isinstance(other, Line):
            return self.crossLine(other)
        elif isinstance(other, HalfLine):
            return self.crossHalfLine(other)
        else:
            raise TypeError(
                "The collisions {}|{} are not dealt with.".format(
                    type(self), type(other)
                )
            )

    __or__ = cross

    def crossForm(self, other):
        """Return the bool: (2 sides are crossing)."""
        points = []
        for s1 in self.sides:
            for s2 in other.sides:
                point = s1.crossSegment(s2)
                if point:
                    points.append(point)
        return points

    def crossDirection(self, other):
        """Return the list of the points of intersection between the form and a segment or a line."""
        points = []
        for side in self.sides:
            cross = side | other
            if cross:
                points.append(cross)
        return points

    def crossHalfLine(self, other):
        """Return the list of points of intersection in order between the form and a half line."""
        points = []
        for segment in self.segments:
            cross = other.crossSegment(segment)
            if cross:
                points.append(cross)
        hp = other.point
        objects = [(p, Point.distance(p, hp)) for p in points]
        objects = sorted(objects, key=lambda x: x[1])
        return [p for (p, v) in objects]

    def crossLine(self, other):
        """Return the list of the points of intersection between the form and a line."""
        points = []
        for segment in self.segments:
            cross = segment.crossLine(other)
            if cross:
                points.append(cross)
        return points

    def crossSegment(self, other):
        """Return the list of the points of intersection between the form and a segment."""
        points = []
        for side in self.sides:
            point = side.crossSegment(other)
            if point:
                points.append(point)
        return points

    def crossSelf(self, e=1e-10):
        """Return the list of the points of intersections between the form and itself."""
        results = []
        l = len(self.segments)
        for i in range(l):
            for j in range(i + 1, l):
                point = self.segments[i].crossSegment(self.segments[j])
                if point:
                    if point in self.points:
                        results.append(point)
        return results

    def convex(self):
        """Return the bool (the form is convex)."""
        x, y = self.center
        angles = []
        l = len(self.points)
        for i in range(l - 1):
            A = self.points[(i + l - 1) % l]
            B = self.points[i % l]
            C = self.points[(i + 1) % l]
            u = Vector.createFromTwoPoints(A, B)
            v = Vector.createFromTwoPoints(C, B)
            angle = v ^ u
            if angle > pi:
                return True
        return False

    def getSparse(
        self,
    ):  # as opposed to makeSparse which keeps the same form and return nothing
        """Return the form with the most sparse points."""
        return copy.deepcopy(self.makeSparse())

    def makeSparse(self):
        """Change the form into the one with the most sparse points."""
        center = self.center
        l = []
        for point in self.points:
            angle = Vector.createFromTwoPoints(point, center).angle
            l.append((angle, copy.deepcopy(point)))
        l = sorted(l, key=lambda x: x[0])
        for i in range(len(l)):
            self.points[i].set(l[i][1])

    def __contains__(self, point):
        """Return the boolean: (the point is in the form)."""
        h = HalfLine(point, 0)
        ps = self.crossHalfLine(h)
        return len(ps) % 2 == 1

    def rotate(self, angle, point=None):
        """Rotate the form by rotating its points from the center of rotation.
        Use center of the shape as default center of rotation."""
        # Actually not working
        if not point:
            point = self.center
        for i in range(len(self.points)):
            self.points[i].rotate(angle, point)

    def move(self, step):
        """Move the object by moving all its points using step."""
        for point in self.points:
            l = min(len(step), len(point.position))
            for i in range(l):
                point.position[i] = step[i]

    def addPoint(self, point):
        """Add a point to the form."""
        self.points.append(point)

    def addPoints(self, points):
        """Add points to the form."""
        self.points.extend(points)

    append = addPoint
    extend = addPoints

    def removePoint(self, point):
        """Remove a point to the form."""
        self.points.remove(point)

    __remove__ = removePoint

    def __getitem__(self, index):
        """Return the point of index index."""
        return self.points[index]

    def __setitem__(self, index, value):
        """Change the points of a form."""
        self.points[index] = value

    @property
    def perimeter(self):
        """Return the perimeter of the form."""
        return sum([s.length for s in self.segments])

    @property  # This can only be a getter
    def area(self):
        """Return the area of the form using its own points.
        General case in 2d only for now..."""
        l = len(self.points)
        if (
            l < 3
        ):  # The form has no point, is a single point or a segment, so it has no area.
            return 0
        elif l == 3:  # The form is a triangle, so we can calculate its area.
            a, b, c = [Vector.createFromSegment(segment) for segment in self.sides]
            A = (
                1
                / 4
                * sqrt(
                    4 * a.norm ** 2 * b.norm ** 2
                    - (a.norm ** 2 + b.norm ** 2 - c.norm ** 2) ** 2
                )
            )
            return A
        else:  # The form has more points than 3, so we can cut it in triangles.
            area = 0
            C = self.center
            for i in range(l):
                A = self.points[i]
                B = self.points[(i + 1) % l]
                triangle = Form([A, B, C])
                area += triangle.area
            return area

    @property
    def angles(self):
        ags = []
        vectors = self.vectors
        ag = (-vectors[-1]).angle
        for v in vectors:
            ang = v.angle - ag
            ag = (-v).angle
            ags.append(ang)
        return ags

    @property
    def abs_angles(self):
        ags = []
        vectors = self.vectors
        ag = (-vectors[-1]).angle
        for v in vectors:
            dag = abs(v.angle - ag)
            ang = min(dag, 2 * math.pi - dag)
            ag = (-v).angle
            ags.append(ang)
        return ags

    def getCircleDiameter(self):
        """Return the circle which diameter is the largest segment of the form.."""
        segments = []
        for (i, p1) in enumerate(self.points):
            for p2 in self.points[i + 1 :]:
                segments.append(Segment(p1, p2))
        s = max(segments, key=lambda s: s.length)
        return Circle(*s.middle, radius=s.length / 2)

    def getBornCircle(self):
        distances = []
        for p1 in self.points:
            distance = sum([Point.distance(p1, p2) for p2 in self.points])
            distances.append((distance, p1))
        pt1 = max(distances)[1]
        pt2 = Point.farthest(pt1, self.points)
        distances = []
        for p in self.points:
            if not p in [pt1, pt2]:
                distance = max(Point.distance(p, pt1), Point.distance(p, pt2))
                distances.append((distance, p))
        pt3 = max(distances)[1]
        triangle = Form([pt1, pt2, pt3])
        return triangle.getBornCircleTriangle()

    def getBornCircle2(self):
        if len(self.points) == 3:
            return self.getBornCircleTriangle()
        else:
            circles = []
            for (p1, p2, p3) in itertools.combinations(self.points, 3):
                form = Form([p1, p2, p3])
                circle = form.getBornCircle2()
                circles.append(circle)
            circles.sort(key=lambda circle: circle.radius, reverse=True)
            return circles[0]

    def getBornCircle3(self):
        triangles = self.getSubForms(3)
        f = max(triangles, key=lambda triangle: triangle.perimeter)
        # f = max(triangles, key=lambda triangle: triangle.area)
        if f.obtusangle:
            return f.getCircleDiameter(), f
        else:
            return f.getCircleCircumscribed(), f

    def getBornCircleSlow(self, e=1e-3):
        for (p1, p2) in self.getSubForms(2):
            segment = Segment(p1, p2)
            l = segment.length / 2
            c = segment.middle
            condition = True
            for p in self.points:
                if Point.distance(p, c) + e > l:
                    condition = False
                    break
            if condition:
                return Circle(*c, radius=l)

        for triangle in self.getSubForms(3):
            circle = triangle.getCircleCircumscribed()
            c = circle.center
            l = circle.radius
            condition = True
            for p in self.points:
                if Point.distance(p, c) + e > l:
                    condition = False
                    break
            if condition:
                print(circle)
                return circle

            print("not found")
            circle.border_color = colors.YELLOW
            return circle

    def getSubForms(self, n):
        return list(map(Form, itertools.combinations(self.points, n)))

    @property
    def obtusangle(self):
        for angle in self.abs_angles:
            if 2 * angle >= math.pi:
                return True
        return False

    @property
    def acutangle(self):
        for angle in self.abs_angles:
            if 2 * angle <= math.pi:
                return True
        return False

    def getCircleCircumscribed(self):
        """Return the circumbscribed circle of a triangle."""
        p1, p2, p3 = self.points
        a1 = -(p2.x - p1.x) / (p2.y - p1.y)
        b1 = (p2.x ** 2 - p1.x ** 2 + p2.y ** 2 - p1.y ** 2) / (2 * (p2.y - p1.y))
        a2 = -(p3.x - p2.x) / (p3.y - p2.y)
        b2 = (p3.x ** 2 - p2.x ** 2 + p3.y ** 2 - p2.y ** 2) / (2 * (p3.y - p2.y))
        x = (b1 - b2) / (a2 - a1)
        y = a1 * x + b1
        radius = math.hypot(p1.x - x, p1.y - y)
        return Circle(x, y, radius=radius)

    def __len__(self):
        """Return number of points."""
        return len(self.points)

    def __xor__(self, other):
        """Return the list of forms that are in the union of 2 forms."""
        if type(other) == Form:
            other = [other]
        return Form.union([other, self])

    def __and__(self, other):
        """Return the list of forms that are in the intersection of 2 forms."""
        points = self.crossForm(other)
        points += [point for point in self.points if point in other]
        points += [point for point in other.points if point in self]
        if points:
            return Form(points)

    # Color
    def setColor(self, color):
        """Color the whole form with a new color."""
        self.point_color = color
        self.side_color = color
        self.area_color = color

    def getColor(self):
        """Return the color of the segments because it is the more widely used."""
        return self.side_color

    def setPointColor(self, color):
        """Set the color of the points of the form."""
        for point in self.points:
            point.color = color

    def getPointColor(self):
        """Return the common color of the points."""
        l = [point.color for point in self.points]
        if l.count(l[0]) == len(l):
            return l[0]
        else:
            raise ValueError(
                "The colors of the points must be the same otherwise it makes no sense."
            )

    def setPointMode(self, mode):
        """Set the mode of the points."""
        for point in self.points:
            point.mode = mode

    def getPointMode(self):
        """Return the common mode of the points."""
        l = [point.mode for point in self.points]
        if l.count(l[0]) == len(l):
            return l[0]
        else:
            raise ValueError(
                "The modes of the points must be the same otherwise it makes no sense."
            )

    def setPointFill(self, fill):
        """Set the fill of the points."""
        for point in self.points:
            point.fill = fill

    def getPointFill(self):
        """Return the common fill attribute of the points."""
        l = [point.fill for point in self.points]
        if l.count(l[0]) == len(l):
            return l[0]
        else:
            raise ValueError(
                "The fill attributes of the points must be the same otherwise it makes no sense."
            )

    def setPointRadius(self, radius):
        """Set the radius of the points."""
        for point in self.points:
            point.radius = radius

    def getPointRadius(self):
        """Return the common radius of the points."""
        l = [point.radius for point in self.points]
        if l.count(l[0]) == len(l):
            return l[0]
        else:
            raise ValueError(
                "The radiuses of the points must be the same otherwise it makes no sense."
            )

    def setPointSize(self, size):
        """Set the size of the points."""
        for point in self.points:
            point.size = size

    def getPointSize(self):
        """Return the common size of the points."""
        l = [point.size for point in self.points]
        if l.count(l[0]) == len(l):
            return l[0]
        else:
            raise ValueError(
                "The sizes of the points must be the same otherwise it makes no sense."
            )

    def setPointWidth(self, width):
        """Set teh width of the points."""
        for point in self.points:
            point.width = width

    def getPointWidth(self):
        """Return the common width of the points."""
        l = [point.width for point in self.points]
        if l.count(l[0]) == len(l):
            return l[0]
        else:
            raise ValueError(
                "The widths of the points must be the same otherwise it makes no sense."
            )

    def setSegmentColor(self, color):
        """Set teh color of the segments."""
        for segment in self.segments:
            segment.color = color

    def getSegmentColor(self):
        """Return the common color of the segments."""
        l = [segment.color for segment in self.segments]
        if l.count(l[0]) == len(l):
            return l[0]
        else:
            raise ValueError(
                "The colors of the points must be the same otherwise it makes no sense."
            )

    def setSegmentWidth(self, width):
        """Set the width of the segments."""
        for segment in self.segments:
            segment.width = width

    def getSegmentWidth(self):
        """Return the common width of the segments."""
        l = [segment.width for segment in self.segments]
        if l.count(l[0]) == len(l):
            return l[0]
        else:
            raise ValueError(
                "The widths of the segments must be the same otherwise it makes no sense."
            )

    getCrossPoints = crossSelf

    sides = segments = property(
        getSegments, setSegments, doc="Represents the segments."
    )
    vectors = property(getVectors, setVectors)
    center = property(getCenter, setCenter, doc="Center of the form.")
    centroid = property(getCentroid, setCentroid, doc="Centroid of the form.")
    color = property(getColor, setColor, doc="Segment color.")
    cross_points = property(
        getCrossPoints, doc="Points of intersection of the sides of the form itself."
    )
    point_color = property(
        getPointColor, setPointColor, doc="Set the color of the points."
    )
    point_mode = property(getPointMode, setPointMode, doc="Mode of the points.")
    point_fill = property(
        getPointFill,
        setPointFill,
        doc="Fill attribute of the circles representing the points.",
    )
    point_radius = property(
        getPointRadius,
        setPointRadius,
        doc="Radius of the circles representing the point.",
    )
    point_size = property(
        getPointSize, setPointSize, doc="Size of the cross that represents the point."
    )
    point_width = property(
        getPointWidth,
        setPointWidth,
        doc="Width of the cross that represents the point.",
    )
    segment_color = property(
        getSegmentColor, setSegmentColor, doc="Color of the segments."
    )
    segment_width = property(
        getSegmentWidth, setSegmentWidth, doc="Width of the segments."
    )


class Circle:
    @classmethod
    def random(cls, borns=[-1, 1], radius_borns=[0, 1], **kwargs):
        """Create a random circle."""
        x = random.uniform(*borns)
        y = random.uniform(*borns)
        r = random.uniform(*radius_borns)
        return cls(x, y, radius=r, **kwargs)

    @classmethod
    def createFromPointAndRadius(cls, point, radius, **kwargs):
        """Create a circle from point."""
        return cls(*point, radius=radius, **kwargs)

    def __init__(
        self,
        *args,
        radius,
        fill=False,
        color=colors.WHITE,
        border_color=None,
        area_color=None,
        center_color=None,
        radius_color=None,
        radius_width=1,
        text_color=None,
        text_size=20
    ):
        """Create a circle object using x, y and radius and optional color and width."""
        if len(args) == 1:
            args = args[0]
        self.position = args
        self.radius = radius
        self.fill = fill
        if color:
            if not border_color:
                border_color = color
            if not area_color:
                area_color = color
            if not radius_color:
                radius_color = color
            if not text_color:
                text_color = color
        self.border_color = border_color
        self.area_color = area_color
        self.center_color = center_color
        self.radius_color = radius_color
        self.radius_width = radius_width
        self.text_color = text_color
        self.text_size = text_size

    def __str__(self):
        """Str representation of a circle."""
        return "c(pos:" + str(self.position) + ",rad:" + str(self.radius) + ")"

    def __contains__(self, position):
        """Determine if a point is in a circle."""
        return (
            sum([(c1 - c2) ** 2 for (c1, c2) in zip(self.position, position)])
            <= self.radius
        )

    def getX(self):
        """Return the x component of the circle."""
        return self.position[0]

    def setX(self, value):
        """Set the x component of the circle."""
        self.position[0] = value

    def getY(self):
        """Return the y component of the circle."""
        return self.position[1]

    def setY(self, value):
        """Set the y component of the circle."""
        self.position[1] = value

    def getPoint(self):
        """Return the point that correspond to the center of the circle."""
        return Point(*self.position)

    def setPoint(self, point):
        """Set the center point of the circle by changing the position of the circle."""
        self.position = point.position

    def getR(self):
        """Return the radius."""
        return self.radius

    def setR(self, radius):
        """Set the radius to the given radius."""
        self.radius = radius

    x = property(getX, setX, "Allow the user to manipulate the x component easily.")
    y = property(getY, setY, "Allow the user to manipulate the y component easily.")
    r = property(getR, setR, "Abbreviation of the radius")
    center = point = property(
        getPoint, setPoint, "Allow the user to manipulate the point easily."
    )

    def show(self, window, color=None, border_color=None, area_color=None, fill=None):
        """Show the circle on screen using the window."""
        if color:
            if not area_color:
                area_color = color
            if not border_color:
                border_color = color
        if not border_color:
            border_color = self.border_color
        if not area_color:
            area_color = self.area_color
        if not fill:
            fill = self.fill
        window.draw.circle(
            window.screen, border_color, self.position, self.radius, fill
        )

    def showCenter(self, window, color=None, mode=None):
        """Show the center of the screen."""
        if not color:
            color = self.center_color
        if not mode:
            mode = self.center_mode
        self.center.show(window, mode=mode, color=color)

    def showText(self, window, text, color=None, size=None):
        """Show a text next to the circle."""
        if not color:
            color = self.text_color
        if not size:
            size = self.text_size
        self.center.showText(window, text, color=color, size=size)

    def showRadius(self, window, color=None, width=None):
        """Show the radius of the circle."""
        if not color:
            color = self.radius_color
        if not width:
            width = self.radius_width
        vector = Vector.createFromPolar(self.radius, 0, color=color)
        vector.show(window, self.center, width=width)
        vector.showText(surface, self.center, "radius", size=20)

    def __call__(self, n):
        """Return the main components of the circle."""
        perimeter = 2 * math.pi
        return Point(math.cos(n / perimeter), math.sin(n / perimeter))

    def isCrossingCircle(self, other):
        """Determine if two circles are crossing."""
        vector = Vector.createFromTwoPoints(self.center, other.center)
        return vector.norm < self.radius + other.radius

    def crossCircle(self, other):
        """Return the intersections points of two circles if crossing else
        return None."""
        if self.isCrossingCircle(other):
            s = Segment(self.center, other.center)
            m = s.middle
            n = math.sqrt(self.radius ** 2 - (s.norm / 2) ** 2)
            a = s.angle + math.pi / 2
            v1 = Vector.createFromPolar(n, a)
            v2 = Vector.createFromPolar(n, -a)
            p1 = v1(m)
            p2 = v2(m)
            return [p1, p2]

    def getArea(self):
        """Return the area of a circle using basic geometry."""
        return math.pi * self.radius ** 2

    area = property(getArea)


if __name__ == "__main__":
    from mycontext import Context

    surface = Context(name="Abstract Demonstration", fullscreen=True)

    p1 = Point(10, 0, radius=0.05, color=colors.YELLOW)
    p2 = Point(20, 20, radius=0.05, color=colors.YELLOW)
    # origin=Point(0,0)
    origin = Point.origin()

    l1 = HalfLine(origin, math.pi / 4)
    l2 = Line(p1, math.pi / 2, correct=False)
    s1 = Segment(p1, p2)
    print(Point.null())

    while surface.open:
        # Surface specific commands
        surface.check()
        surface.control()
        surface.clear()
        surface.show()

        # Actions
        l1.rotate(0.01, p2)
        l2.rotate(-0.02, p1)
        s1.rotate(0.03)

        p = l1 | l2

        o = Point(0, 0)
        p3 = l2.projectPoint(o)
        f = Form([p1, p2, p3], area_color=colors.RED, fill=True)

        # Show
        surface.draw.window.print("l1.angle: " + str(l1.angle), (10, 10))
        surface.draw.window.print("l2.angle: " + str(l2.angle), (10, 30))
        surface.draw.window.print("f.area: " + str(f.area), (10, 50))

        f.show(surface)
        f.center.show(surface)

        s1.show(surface)

        o.show(surface, color=colors.GREY)
        o.showText(surface, "origin")

        p3.showText(surface, "origin's projection")
        p3.show(surface, color=colors.LIGHTGREY)

        if p:
            p.show(surface, color=colors.RED)
            p.showText(surface, "intersection point", color=colors.RED)

        p1.show(surface)
        p1.showText(surface, "p1")

        p2.show(surface)
        p2.showText(surface, "p2")

        l1.show(surface, color=colors.GREEN)
        l1.point.show(surface, color=colors.LIGHTGREEN, mode="cross", width=3)
        l1.vector.show(surface, l1.point, color=colors.LIGHTGREEN, width=3)

        l2.show(surface, color=colors.BLUE)
        l2.point.show(surface, color=colors.LIGHTBLUE, mode="cross", width=3)
        l2.vector.show(surface, l2.point, color=colors.LIGHTBLUE, width=3)

        # Flipping the screen
        surface.flip()



# Show interpolations being made while moving the points.

# Imports
from interpolation import PolynomialInterpolation
from curves import Trajectory
from context import Context
from abstract import Point
import mycolors
import random

context = Context()
l = 10
points = [Point(2 * x, random.randint(-5, 5)) for x in range(l)]
n = 0
ncp = 200  # number construction points

while context.open:
    context.check()
    context.control()
    context.clear()
    context.show()

    n = (n + 1) % (ncp + 1)
    Point.turnPoints([1 / 1000 for i in range(l)], points)

    pi1 = PolynomialInterpolation(points)
    pi2 = PolynomialInterpolation(points)
    # pi2.createPolynomialsRespectingDistance()
    ti = Trajectory(points)
    p1 = Point(*pi1(n / ncp))
    p2 = Point(*pi2(n / ncp))
    p3 = Point(*ti(n / ncp))

    # l1=Line.createFromTwoPoints(p1,p2)

    pi1.show(context, n=200)
    ti.show(context)
    p1.show(context, color=mycolors.YELLOW, radius=0.1, fill=True)
    p2.show(context, color=mycolors.YELLOW, radius=0.1, fill=True)
    p3.show(context, color=mycolors.YELLOW, radius=0.1, fill=True)

    context.flip()



from context import Context
from pygame.locals import *
import pygame


class SimpleManager:
    """Manage a program using the context by many having functions that can be
    overloaded to make simple and fast programs.
    This process allow the user to think directly about what the program does,
    instead of focusing on how to display things.
    The way it works is by making the main class of the program inheriting from
    this one."""

    def __init__(self, name="SimpleManager", **kwargs):
        """Create a context manager with the optional name."""
        self.context = Context(name=name, **kwargs)
        self.pause = False

    def __call__(self):
        """Call the main loop."""
        while self.context.open:
            self.events()
            if not self.pause:
                self.update()
            self.show()

    def events(self):
        """Deal with all the events."""
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                self.context.open = False
            if event.type == KEYDOWN:
                if event.key == K_ESCAPE:
                    self.context.open = False
                if event.key == K_SPACE:
                    self.pause = not (self.pause)
                if event.key == K_f:
                    self.context.switch()  # Set or reverse fullscreen
            if event.type == MOUSEBUTTONDOWN:
                if event.button == 4:
                    self.context.draw.plane.zoom([1.1, 1.1])
                if event.button == 5:
                    self.context.draw.plane.zoom([0.9, 0.9])

    def update(self):
        """Update the context manager."""

    def show(self):
        """Show the context manager."""
        self.context.refresh()
        self.context.control()
        self.context.flip()


class oldManager:  # This manager is deprecated
    """Manage a program using the context by many having functions that can be
    overloaded to make simple and fast programs.
    This process allow the user to think directly about what the program does,
    instead of focusing on how to display things.
    The way it works is by making the main class of the program inheriting from
    this one."""

    def __init__(self, context):
        """Create a context manager."""
        self.context = context

    def __call__(self):
        """Call the main loop."""
        self.main()

    def main(self):
        """Execute the main loop."""
        self.loop()

    def loop(self):
        """Main loop."""
        while self.context.open:
            self.perform()

    def perform(self):
        """Execute the content of the main loop."""
        self.detect()
        self.update()
        self.show()

    def debug(self):
        self.context.console(str(ContextManager.__dict__["react"].__dict__))

    def update(self):
        """Update the context manager."""
        pass

    def show(self):
        """Show the context manager."""
        self.context.refresh()
        self.context.control()
        self.debug()

        self.context.flip()

    def detect(self):
        """Deal with all the events."""
        for event in pygame.event.get():
            self.react(event)

    def react(self, event):
        """React to a given event."""
        self.reactToQuit(event)
        if event.type == KEYDOWN:
            self.reactToKeyboard(event)
            return

    def reactToQuit(self, event):
        """Close the context if the quit button is pressed."""
        if event.type == pygame.QUIT:
            self.context.open = False

    def reactToKeyboard(self):
        """React to the keyboards buttons being pressed."""
        self.reactToEscape()

    def reactToEscape(self):
        """React to the button escape."""
        if event.key == K_ESCAPE:
            self.context.open = False


class Manager:
    def __init__(self, name="Manager", dt=10e-3, **kwargs):
        """Create a manager using a context, this methods it to be overloaded."""
        self.context = Context(name=name, **kwargs)
        self.count = self.context.count
        self.pause = False
        self.dt = dt
        # Typing stuff
        self.typing = False
        self.alphabet = "abcdefghijklmnopqrstuvwxyz"
        self.caps_numbers = ")!@#$%^&*("
        self.numbers = "0123456789"
        self.typing = False
        self.shiftlock = False
        self.capslock = False
        self.altlock = False

    def __str__(self):
        """Return the string representation of the manager."""
        return type(self).__name__ + "(\n{}\n)".format(
            "\n".join(map(lambda x: ":".join(map(str, x)), self.__dict__.items()))
        )

    def __call__(self):
        """Call the main loop, this method is to be overloaded."""
        self.main()

    def main(self):
        """Main loop of the simple manager."""
        self.setup()  # Name choices inspired from processing
        while self.context.open:
            self.loop()

    def setup(self):
        """Code executed before the loop."""
        pass

    def loop(self):
        """Code executed during the loop."""
        self.eventsLoop()
        self.updateLoop()
        self.showLoop()

    def eventsLoop(self):
        """Deal with the events in the loop."""
        for event in pygame.event.get():
            self.react(event)

    def react(self, event):
        """React to the pygame events."""
        if event.type == QUIT:
            self.switchQuit()
        elif event.type == KEYDOWN:
            self.reactKeyDown(event.key)
        elif event.type == KEYUP:
            self.reactKeyUp(event.key)
        elif event.type == MOUSEBUTTONDOWN:
            self.reactMouseButtonDown(event.button, event.pos)
        elif event.type == MOUSEBUTTONUP:
            self.reactMouseButtonUp(event.button, event.pos)
        elif event.type == MOUSEMOTION:
            self.reactMouseMotion(event.pos)

    def switchQuit(self):
        """React to a quit event."""
        self.context.open = not (self.context.open)

    def reactKeyDown(self, key):
        """React to a keydown event."""
        self.reactAlways(key)
        if self.typing:
            self.reactTyping(key)
        else:
            self.reactMain(key)

    def reactKeyUp(self, key):
        """React to a keyup event."""
        pass

    def reactAlways(self, key):
        """React to a key whether or not the typing mode is on."""
        # print(key) for debugging the keys
        if key == K_ESCAPE:
            self.switchQuit()
        if key == K_SLASH or key == K_BACKSLASH:
            if not self.typing:
                self.context.console("Typing activated.")
            self.typing = True
        if key == K_BACKQUOTE:
            self.switchTyping()

    def reactLock(self, key):
        """React to a locking key."""
        if key == K_CAPSLOCK:
            self.capslock = not (self.capslock)
        elif key == K_LSHIFT or key == K_RSHIFT:
            self.shiftlock = True
        elif key == K_LALT or key == K_RALT:
            self.altlock = True

    def reactTyping(self, key):
        """React to a typing event."""
        self.reactLock(key)
        if self.altlock:
            self.reactAltCase(key)
        elif self.capslock or self.shiftlock:
            self.reactUpperCase(key)
        else:
            self.reactLowerCase(key)

        if key == K_SPACE:
            self.write(" ")
        elif key == 8:
            self.delete()
        if key == K_LCTRL:
            self.context.console.nextArg()
        elif key == K_UP:
            self.context.console.back()
        elif key == K_DOWN:
            self.context.console.forward()
        elif key == K_RETURN:
            self.eval()
            self.context.console.nextLine()

    def eval(self):
        """Execute a line."""
        content = self.context.console.line.content
        if content[0] == "/":
            for command in content[1:]:
                try:
                    self.context.console(str(eval(command)))
                except:
                    self.context.console("Invalid command.")
        if content[0] == "\\":
            for command in content[1:]:
                try:
                    exec(command)
                    self.context.console("Command " + command + " executed.")
                except Exception as e:
                    self.context.console(str(e))
        self.context.console.eval()

    def reactAltCase(self, key):
        """React when typing with alt key pressed."""
        if key == K_e:
            self.write("`")  # Stupid
        elif key == 167:
            self.write("´")

    def reactLowerCase(self, key):
        """React when typing in lower case."""
        d = {
            K_COMMA: ",",
            K_PERIOD: ".",
            K_SEMICOLON: ";",
            K_LEFTBRACKET: "[",
            K_RIGHTBRACKET: "]",
            39: "'",
            45: "-",
            K_EQUALS: "=",
        }
        if 48 <= key <= 57:
            self.write(self.numbers[key - 48])
        elif 97 <= key <= 122:
            self.write(self.alphabet[key - 97])
        elif key in d:
            self.write(d[key])
        elif key == K_SLASH:
            if not self.context.console.line.empty:
                self.context.console.nextLine()
            self.write("/")
            self.context.console.nextArg()
        elif key == K_BACKSLASH:
            if not self.context.console.line.empty:
                self.context.console.nextLine()
            self.write("\\")
            self.context.console.nextArg()

    def reactUpperCase(self, key):
        """React to a key when typing in uppercase."""
        d = {59: ":''", 44: "<", 46: ">", 47: "?", 45: "_", 39: '"', 61: "+"}
        if 48 <= key <= 57:
            self.write(self.caps_numbers[key - 48])
        elif 97 <= key <= 122:
            self.write(self.alphabet[key - 97].upper())
        elif key in d:
            self.write(d[key])

    def write(self, c):
        """Write some content."""
        self.context.console.lines[-1].content[-1] += c
        self.context.console.lines[-1].refresh()
        self.shiftlock = False
        self.altlock = False

    def delete(self, n=1):
        """Delete some content."""
        self.context.console.lines[-1].content[-1] = self.context.console.lines[
            -1
        ].content[-1][:-n]
        self.context.console.lines[-1].refresh()

    def reactMain(self, key):
        """React as usual when not typing."""
        if key == K_f:
            self.switchFullscreen()
        if key == K_1:
            self.switchCapture()
        if key == K_2:
            self.switchCaptureWriting()
        if key == K_3:
            self.switchScreenWriting()
        if key == K_LALT:
            self.switchPause()

    def switchTyping(self):
        """Switch the typing mode."""
        self.typing = not (self.typing)
        if self.typing:
            self.context.console("Typing activated.")
            self.context.console.nextLine()
        else:
            self.context.console("Typing deactivated.")

    def switchScreenWriting(self):
        """Switch the screen writing mode."""
        if self.context.camera.screen_writing:
            self.context.camera.screen_writer.release()
        self.context.camera.switchScreenWriting()
        if self.context.camera.screen_writing:
            self.context.console("The screen is being written.")
        else:
            self.context.console("The screen video has been released")
            self.context.console("and is not being written anymore.")

    def switchCaptureWriting(self):
        """Switch the capture writing mode."""
        if self.context.camera.capture_writing:
            self.context.camera.capture_writer.release()
        self.context.camera.switchCaptureWriting()
        if self.context.camera.capture_writing:
            self.context.console("The capture is being written.")
        else:
            self.context.console("The capture video has been released")
            self.context.console("and is not being written anymore.")

    def switchPause(self):
        """React to a pause event."""
        self.pause = not self.pause
        if self.pause:
            self.context.console("The system is paused.")
        else:
            self.context.console("The system is unpaused.")

    def switchCapture(self):
        """React to a capture event."""
        self.context.camera.switchCapture()
        if self.context.camera.capturing:
            self.context.console("The camera capture is turned on.")
        else:
            self.context.console("The camera capture is turned off.")

    def switchFullscreen(self):
        """React to a fullscreen event."""
        self.context.switch()
        if self.context.fullscreen:
            self.context.console("The fullscreen mode is set.")
        else:
            self.context.console("The fullscreen mode is unset.")

    def reactMouseButtonDown(self, button, position):
        """React to a mouse button down event."""
        if button == 4:
            self.context.draw.plane.zoom([1.1, 1.1])
        if button == 5:
            self.context.draw.plane.zoom([0.9, 0.9])

    def reactMouseButtonUp(self, button, position):
        """React to a mouse button up event."""
        pass

    def reactMouseMotion(self, position):
        """React to a mouse motion event."""
        pass

    def updateLoop(self):
        """Update the manager while in the loop."""
        if not self.pause:
            self.update()
            self.count()
        self.context.camera.write()  # Write on the camera writers if they are on

    def update(self):
        """Update the components of the manager of the loop. This method is to be
        overloaded."""
        pass

    def showLoop(self):
        """Show the graphical components and deal with the context in the loop."""
        if not self.typing:  # Ugly fix for easier pratical use
            self.context.control()
        self.context.clear()
        self.context.show()
        self.show()
        self.showCamera()
        self.context.console.show()
        self.context.flip()

    def show(self):
        """Show the graphical components on the context. This method is to be
        overloaded."""
        pass

    def showCamera(self):
        """Show the camera if active."""
        if self.context.camera.capturing:
            self.context.camera.show()

    def counter():
        doc = "The Counter property."

        def fget(self):
            """Bind the counter of the manager to the one of the context."""
            return self.context.counter

        def fset(self, counter):
            """Set the counter of the context."""
            self.context.counter = counter

        return locals()

    counter = property(**counter())


class BodyManager(Manager):
    """Manage the bodies that are given when initalizing the object."""

    @classmethod
    def createRandomBodies(cls, t, n=5, **kwargs):
        """Create random bodies using their class t.
        The bodies of the class t must have a random method."""
        bodies = [t.random() for i in range(n)]
        return cls(bodies, **kwargs)

    def __init__(self, *bodies, following=False, **kwargs):
        """Create a body manager using its bodies and optional arguments for the context."""
        super().__init__(**kwargs)
        self.bodies = bodies
        self.following = following

    def update(self):
        """Update the bodies."""
        self.updateFollowers()
        self.updateBodies()

    def updateFollowers(self):
        """Update the bodies that are followers."""
        # This is not implemented correctly for now."""
        p = self.context.point()
        if self.following:
            for body in self.bodies:
                body.follow(p)

    def updateBodies(self):
        """Update the bodies individually."""
        for body in self.bodies:
            body.update(self.dt)

    def show(self):
        """Show the bodies."""
        self.showBodies()

    def showBodies(self):
        """Show all the bodies individually."""
        for body in self.bodies:
            body.show(self.context)


class EntityManager(Manager):
    """Create a manager that deals with entities."""

    def __init__(self, *entities, dt=0.1, friction=0.1, **kwargs):
        """Create an entity manager using the list of entities and optional
        arguments for the manager."""
        super().__init__(**kwargs)
        self.entities = list(entities)
        self.dt = dt
        self.friction = friction

    def update(self):
        """Update all entities."""
        for entity in self.entities:
            entity.update(self.dt)

    def show(self):
        """Show all entities."""
        for entity in self.entities:
            entity.show(self.context)

    def reactKeyDown(self, key):
        """Make all entities react to the keydown event."""
        super().reactKeyDown(key)
        for entity in self.entities:
            entity.reactKeyDown(key)

    def reactMouseMotion(self, position):
        """Make all entities react to the mouse motion."""
        position = self.context.getFromScreen(tuple(position))
        for entity in self.entities:
            entity.reactMouseMotion(position)

    def reactMouseButtonDown(self, button, position):
        """Make all entities react to the mouse button down event."""
        position = self.context.getFromScreen(tuple(position))
        for entity in self.entities:
            entity.reactMouseButtonDown(button, position)

    def spread(self, n=10):
        """Spread randomly the bodies."""
        for entity in self.entities:
            entity.motion *= n

    def setFriction(self, friction):
        """Set the friction of all entities to the given friction."""
        for entity in self.entities:
            entity.setFriction(friction)


class BlindManager(Manager):
    """Ugly way to make a manager without camera."""

    def __init__(self, camera=False, **kwargs):
        super().__init__(camera=camera, **kwargs)


class AbstractManager(Manager):
    """Manager that deals with abstract objects."""

    def __init__(self, *group, **kwargs):
        """Create a group of abstract (geometrical) objects."""
        super().__init__(**kwargs)
        self.group = list(group)

    def show(self):
        """Show the objects of the group."""
        for e in self.group:
            e.show(self.context)


class GameManager(Manager):
    """Base class Manager for games."""

    def __init__(self, game, controller=None, **kwargs):
        super().__init__(**kwargs)
        self.game = game
        self.controller = controller
        self.context.units = [10, 10]

    def update(self):
        if self.controller:
            player = self.getPlayer()
            if player and not self.game.won:
                self.context.position = player.position
        self.game.update()

    def showLoop(self):
        """Show the graphical components and deal with the context in the loop."""
        if not self.typing:  # Ugly fix for easier practical use
            # self.context.control()
            pass
        if self.game.won:
            self.context.control()
        self.context.clear()
        self.show()
        self.showCamera()
        self.context.console.show()
        self.context.flip()

    def show(self):
        self.game.show(self.context)

    def reactKeyDown(self, key):
        super().reactKeyDown(key)
        self.game.reactKeyDown(key)

    def reactMouseMotion(self, position):
        position = self.context.getFromScreen(position)
        self.game.reactMouseMotion(position)

    def reactMouseButtonDown(self, button, position):
        position = self.context.getFromScreen(position)
        self.game.reactMouseButtonDown(button, position)

    def getPlayer(self):
        return self.game.control(self.controller)


class ActivityManager(Manager):
    """Activity manager inspired from android studio"""

    def __init__(self, activities, **kwargs):
        super().__init__(**kwargs)
        self.activities = activities
        self.index = 0

    def show(self):
        self.activity.show(self.context)

    def update(self):
        self.activity.update(self.dt)

    def reactKeyUp(self, key):
        self.activity.onKeyUp(key)

    def reactKeyDown(self, key):
        self.activity.onKeyDown(key)

    def reactMouseButtonDown(self, button, position):
        position = self.context.getFromScreen(position)
        self.activity.onMouseButtonDown(button, position)

    def reactMouseMotion(self, position):
        position = self.context.getFromScreen(position)
        self.activity.onMouseMotion(position)

    def getActivity(self):
        return self.activities[self.index]

    def setActivity(self, activity):
        self.activities[self.index] = activity

    activity = property(getActivity, setActivity)


if __name__ == "__main__":
    from myentity import Entity

    m = EntityManager.random(build=False)
    print(m)
    # cm()



import time


def timer(function):
    """Print the duration of a function making use of decorators."""

    def function_(*args, **kwargs):
        """Tested function."""
        ti = time.time()
        result = function(*args, **kwargs)
        tf = time.time()
        dt = tf - ti
        print("[TIMER]: " + str(function.__name__) + " took " + str(dt) + " seconds.")
        return result

    return function_



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fourier_drawing']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.22.4,<2.0.0', 'opencv-python>=4.5.5,<5.0.0', 'pygame>=2.1.2,<3.0.0']

setup_kwargs = {
    'name': 'fourier-drawing',
    'version': '0.1.5',
    'description': '',
    'long_description': "# Use fourier transform to draw epicycles with your drawings.\n\nDraw a picture or sample one, press space, then watch an epic simulation of epicycles being drawned identically as your picture. Thanks the the fourier transform your drawing will be reproduced in a real-time simulation only using epicycles.\n\n# Demo\n\n![https://www.youtube.com/watch?v=86bYtJCwQ_o](https://cdn.discordapp.com/attachments/507519157387132940/808039024022257694/fourier.gif)\n# Install\n\n```bash\n#Clone the repository\ngit clone git@github.com:MarcPartensky/Fourier.git\ncd Fourier\n\n#Install requirements\npip install -r requirements.txt\n```\n\n# Usage\n\nPut your model image `image.png` in the `FourierImages` folder.\n\n* Option 1: Give image at launch.\n\n```bash\npython __main__.py image.png\n```\n\n* Option 2: Launch then give the image.\n\n```bash\npython __main__.py\n> image name:\n```\n\nThen give your image:\n\n```bash\n> image name:image.png\n```\n\n# Run with docker\n```\n# Replace the first occurence of 'FourierImages' by your custom image folder\n# Replace 'rodolphe.jpg' by your own image too\ndocker-compose run \\\n    -v ./FourierImages:/opt/FourierImages \\\n    --entrypoint './__main__.py rodolphe.jpg' \\\n    fourier\n```\n\n# Description\n\nThere are 3 modes in this program:\n\n* Mode 1: **Sampling**\nSample or draw a picture.\n\n* Mode 2: **Drawing**\nWatch the epicycles simulation which uses fourier transform.\n\n* Mode 3: **Display**\nGet the output image directly without waiting for the simulation.\n\n# Controls\n\n* `Space`: Switch to next mode.\n* `Enter`: Go back to the center.\n* `Up/Down/Right/Left Arrow`: Move arround.\n* `Right/Left Shift`: Zoom in or out.\n* `Quit/Escape`: Quit.\n* `Z`: Cancel last sample.\n* `R`: Remove all samples.\n* `S`: Save the fourier-coefficients.\n\n## Hide or Show the graphical components\nPress the following numbers to toggle:\n* `1`: Image\n* `2`: Green lines\n* `3`: Red graph\n* `4`: White vectors\n* `5`: Grey circles\n* `6`: Yellow sample\n\n# Enjoy!\n",
    'author': 'Marc Partensky',
    'author_email': 'marc.partensky@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from .window import Window
from pygame.locals import *

import math
import numpy as np
from . import colors


class Plane:
    def __init__(self, theme={}, view=None):
        """Create a plane using optionals theme and view."""
        self.createTheme(theme)
        self.createView(view)

    def createTheme(self, theme={}):
        """Initializes the position and the colors for the view of the plane using optional theme."""
        if not "background" in theme:
            theme["background"] = (0, 0, 0)
        if not "grid color" in theme:
            theme["grid color"] = (20, 20, 20)
        if not "grid nscale" in theme:
            theme["grid nscale"] = 3
        if not "show scale" in theme:
            theme["show scale"] = False
        if not "show origin" in theme:
            theme["show origin"] = False
        if not "show units" in theme:
            theme["show units"] = False
        self.theme = theme

    def createView(self, view=None):
        """Initializes the position and the units for the view of the plane using optional view."""
        """View must be a list of length 2."""
        if view:
            self.position = view[0]
            self.units = view[1]
            self.default_position = view[0][:]
            self.default_units = view[1][:]
        else:
            self.default_position = [
                0,
                0,
            ]  # position of the center of the view in the plane's coordonnates
            self.default_units = [40, 40]  # units of the conversion from window/plane
            self.position = [0, 0]
            self.units = [
                40,
                40,
            ]  # units can be handled [-10**-14,10**307] before bugging or crashing

    def __call__(
        self, window
    ):  # The main loop must be redefined by the client and not the functions called within it.
        """Main loop of the plane."""
        window.rename("Plane")
        while window.open:
            window.check()
            self.control(window)
            self.clear(window)
            self.show(window)
            self.flip(window)

    def contains(self, position, window):
        """Determine if a position is in the visible screen."""
        x, y = position
        xmin, ymin, xmax, ymax = self.getCorners(window)
        return (xmin <= x <= xmax) and (ymin <= y <= ymax)

    def flip(self, window):
        """Flip the plane's window."""
        window.flip()

    def clear(self, window, color=None):
        """Clear the plane."""
        if not color:
            color = self.theme["background"]
        window.clear(color)

    def control(self, window):
        """Control the view of the plane."""
        self.controlZoom(window)
        self.controlPosition(window)

    def controlZoom(self, window):
        """Control the zoom of the plane."""
        keys = window.press()
        ux, uy = self.units
        if keys[K_RSHIFT]:
            self.zoom([1.1, 1.1])
        if keys[
            K_LSHIFT
        ]:  # and (ux>2 and uy>2): #The second condition is useless but prevent the user from watching errors due to too far zooming out.
            self.zoom([0.9, 0.9])

    def controlPosition(self, window):
        """Control the position of the plane."""
        keys = window.press()
        if keys[K_RETURN]:
            self.units = self.default_units[:]
            self.position = self.default_position[:]
        ux, uy = self.units
        wsx, wsy = window.size
        x, y = self.position
        k = 50  # Allow the user to move in the grid a reasonable velocity.
        if keys[K_LEFT]:
            x -= wsx / ux / k
        if keys[K_RIGHT]:
            x += wsx / ux / k
        if keys[K_DOWN]:
            y -= wsy / uy / k
        if keys[K_UP]:
            y += wsy / uy / k
        self.position = [x, y]

    def getUnitsColor(self, scale, ascale, color=None):
        """Get the color given a x position."""
        if not color:
            color = self.theme["grid color"]
        dscale = scale - ascale
        return [c * (2 ** dscale) for c in color]

    def show(self, window):
        """Show the elements on screen using the window."""
        self.showGrids(window)

    def showAll(self):
        self.showGrids(window)
        if self.theme["show scale"]:
            window.print("scale:" + str(self.getScale(window)), (10, 10), size=20)
        if self.theme["show origin"] and self.contains((0, 0), window):
            self.showOrigin(window)
        if self.theme["show units"]:
            window.print("units:" + str(self.units), (10, 10), size=20)
        # self.showUnits(window) #Does not work for now

    def showOrigin(self, window, color=colors.WHITE, radius=3):
        """Show the origin of the plane."""
        position = self.getToScreen([0, 0], window)
        window.draw.circle(window.screen, color, position, radius, 0)

    def showUnits(
        self, window
    ):  # Not working because of error of synchronisation between text base system and normal window base.
        """Show the unit of the grid using the window."""
        px, py = self.position
        wsx, wsy = window.size
        ux, uy = self.units
        nx = int(wsx / ux)
        ny = int(wsy / uy)
        for y in range(int(-ny + py), int(ny + py) + 1, 10):
            for x in range(int(-nx + px), int(nx + px) + 1, 10):
                X, Y = self.getToScreen([x, y], window)
                window.print(str([x, y]), [X, Y], size=20)

    def showGrids(self, window, nscale=None):
        """Show the grid using the window and optional scales."""
        ascale = self.getScale(window)  # ascale like actual_scale
        for scale in self.getScales(window, nscale, ascale):
            self.showGrid(window, scale)

    def getScales(self, window, nscale=None, scale=None):
        """Return the scales ."""
        if not nscale:
            nscale = self.theme["grid nscale"]
        if not scale:
            scale = self.getScale(window)
        l = [scale - i for i in range(nscale)]
        l.reverse()
        return l

    def getScale(self, window):
        """Return the actual scale of the displayed screen."""
        ux, uy = self.units
        wsx, wsy = window.size
        uwx, uwy = wsx / ux, wsy / uy
        uw = min(uwx, uwy)
        return int(math.log(uw) / math.log(10))

    def showGrid(self, window, scale=0, ascale=None):
        """Show the grid using the window."""
        if not ascale:
            ascale = self.getScale(window)
        xmin, ymin, xmax, ymax = self.getCorners(window)  # Get the corners of the plane
        # For each line find the begining and the end in the plane's coordonnates then convert it into screen's coordonnates.
        for x in self.arange(xmin, xmax, scale):
            color = self.getUnitsColor(scale, ascale)
            start = self.getToScreen([x, ymin], window)
            end = self.getToScreen([x, ymax], window)
            window.draw.line(window.screen, color, start, end, 1)
        # Repeat the process for the y component
        for y in self.arange(ymin, ymax, scale):
            color = self.getUnitsColor(scale, ascale)
            start = self.getToScreen([xmin, y], window)
            end = self.getToScreen([xmax, y], window)
            window.draw.line(window.screen, color, start, end, 1)

    def arange(self, min, max, scale):
        """Return the list of values that are between min and max, separated by unit."""
        min = np.around(min, -scale)
        max = np.around(max, -scale)
        return np.arange(min, max, 10 ** scale)

    def zoom(self, zoom):
        """Allow the user to zoom into the plane."""
        for i in range(2):
            self.units[i] *= zoom[i]

    def xZoom(self, zoom):
        """Allow the user to zoom into the plane according to the x component."""
        self.units[0] *= zoom

    def yZoom(self, zoom):
        """Allow the user to zoom into the plane according to the y component."""
        self.units[1] *= zoom

    def getToScreen(self, position, window):
        """Return a screen position using a position in the plane."""
        x, y = position
        px, py = self.position
        ux, uy = self.units
        wsx, wsy = window.size
        x = int((x - px) * ux + wsx / 2)
        y = int(wsy / 2 - (y - py) * uy)
        return [x, y]

    def getAllToScreen(self, positions, window):
        """Return the list of screen positions using the list of plane positions."""
        screen_positions = []
        for position in positions:
            screen_positions.append(self.getToScreen(position, window))
        return screen_positions

    def getFromScreen(self, position, window):
        """Return a plane position using a position in the screen."""
        x, y = position
        px, py = self.position
        ux, uy = self.units
        wsx, wsy = window.size
        x = (x - wsx / 2) / ux + px
        y = (wsy / 2 - y) / uy + py
        return [x, y]

    def getAllFromScreen(self, position, window):
        """Return the list of plane positions using the list of screen positions."""
        plane_positions = []
        for position in positions:
            planes_positions.append(self.getFromScreen(position, window))
        return plane_positions

    def getCorners(self, window):
        """Return the corners of the present view."""
        wsx, wsy = window.size
        ux, uy = self.units
        x, y = self.position
        lx, ly = wsx / ux, wsy / uy
        return [x - lx / 2, y - ly / 2, x + lx / 2, y + ly / 2]

        mx, my = self.getFromScreen([0, wsy], window)
        Mx, My = self.getFromScreen([wsx, 0], window)
        return (mx, my, Mx, My)

    def setCorners(self, corners, window):
        """Change the actual corners of the plane by changing its position and units."""
        xmin, ymin, xmax, ymax = corners
        wsx, wsy = window.size
        self.position = [xmin + (xmax - xmin) / 2, ymin + (ymax - ymin) / 2]
        self.units = [wsx / (xmax - xmin), wsy / (ymax - ymin)]

    # The following functions are static methods

    def getCornersFromCoordonnates(coordonnates):
        """Return the corners (top_left_corner,bottom_right_corner) using the coordonnates (position+size)."""
        """[x,y,sx,sy] -> [mx,my,Mx,My]"""
        x, y, sx, sy = coordonnates
        mx, my = x - sx / 2, y - sy / 2
        Mx, My = x + sx / 2, y + sy / 2
        corners = (mx, my, Mx, My)
        return corners

    def getCoordonnatesFromCorners(corners):
        """Return the coordonnates (position+size) using the corners (top_left_corner,bottom_right_corner)."""
        """[mx,my,Mx,My] -> [x,y,sx,sy]"""
        mx, my, Mx, My = corners
        sx, sy = Mx - mx, My - my
        x, y = mx + sx / 2, my + sy / 2
        coordonnates = (x, y, sx, sy)
        return coordonnates

    def getCoordonnatesFromRect(rect):
        """Return the coordonnates (position,size) using the rect (top_left_corner,size)."""
        """[x,y,sx,sy] -> [mx,my,sx,sy]"""
        mx, my, sx, sy = rect
        x, y = mx + sx / 2, my + sy / 2
        coordonnates = [x, y, sx, sy]
        return coordonnates

    def getRectFromCoordonnates(coordonnates):
        """Return the rect (top_left_corner,size) using the coordonnates (position,size)."""
        """[mx,my,sx,sy] -> [x,y,sx,sy]"""
        x, y, sx, sy = coordonnates
        mx, my = x - sx / 2, y - sy / 2
        rect = [mx, my, sx, sy]
        return rect

    def getRectFromCorners(corners):
        """Return the rect (top_left_corner,size) using the corners (top_left_corner,bottom_right_corner)."""
        """[mx,my,Mx,My] -> [mx,my,sx,sy]"""
        mx, my, Mx, My = corners
        sx, sy = Mx - mx, My - my
        rect = [mx, my, sx, sy]
        return rect

    def getCornersFromRect(rect):
        """Return the (top_left_corner,bottom_right_corner) using the corners rect (top_left_corner,size)."""
        """[mx,my,Mx,My] -> [mx,my,sx,sy]"""
        mx, my, sx, sy = rect
        Mx, My = mx + sx, my + sy
        corners = [mx, my, Mx, My]
        return corners


if __name__ == "__main__":
    window = Window(fullscreen=False)
    theme = {
        "background": (0, 0, 0),
        "show scale": True,
        "show origin": True,
        "grid color": (200, 200, 200),
    }
    plane = Plane(theme)
    # plane.units=[1/10**300,1/10**300]
    plane(window)
